import datetime
import random
import uuid

import tator


def random_state(project, state_type, video_obj, post=False):
    attributes = {}
    out = {
        "project": project,
        "type": state_type,
        "media_ids": [video_obj.id],
        "frame": random.randint(0, video_obj.num_frames - 1),
        "attributes": attributes
    }

    return {**out}


def test_state_type_delete(host, token, project, video_type, video):
    tator_api = tator.get_api(host, token)
    response = tator_api.create_state_type(
        project,
        state_type_spec={
            "name": "state_type",
            "description": "Test state type",
            "project": project,
            "media_types": [video_type],
            "association": "Frame",
            "attribute_types": [],
        },
    )
    state_type = response.id
    video_obj = tator_api.get_media(video)

    # Verify no states exist to start
    response = tator_api.get_state_list(project, type=state_type)
    assert len(response) == 0

    # Test bulk create.
    num_states = random.randint(20, 100)
    states = [random_state(project, state_type, video_obj, post=True) for _ in range(num_states)]
    state_ids = []
    for response in tator.util.chunked_create(tator_api.create_state_list, project, body=states):
        state_ids += response.id
    assert len(state_ids) == len(states)
    print(f"Created {len(state_ids)} states!")

    # Verify list is the right length
    response = tator_api.get_state_list(project, type=state_type)
    assert len(response) == num_states

    response = tator_api.delete_state_type(state_type)

    assert str(state_type) in response.message, "State type id not found in delete response"
    assert str(num_states) in response.message, "State count not found in delete response"
    try:
        caught_it = False
        count = -1
        count = tator_api.get_state_count(project, type=state_type)
    except:
        caught_it=True
    finally:
        assert caught_it or count == 0



from difflib import get_close_matches
from functools import partial
import logging
import re
from typing import Optional


logger = logging.getLogger(__name__)


class Paginator:
    """Manages retrieving pages of results for `get_*_list[_by_id]` API functions"""

    def __init__(self, fn, name, batch_size):
        self._fn = fn
        self._name = name
        self._batch_size = batch_size

    def paginate(self, **kwargs):
        """
        Passes `kwargs` to the stored function and yields pages of results with (at most)
        `_batch_size` results.
        """

        # Set up function with all shared arguments as a partial function
        get_page = partial(self._fn, **kwargs)

        # Initialize progress tracking variables
        page_count = self._batch_size
        page_num = 0
        cum_count = 0
        last_id = None
        msg = self._name + " returned page {} ({} cumulative results)."

        # Loop until a page smaller than the desired `_batch_size` is returned, this signals the end
        # of the list
        while page_count == self._batch_size:
            page = get_page(after=last_id, start=0, stop=self._batch_size)
            page_count = len(page)
            cum_count += page_count
            page_num += 1
            logger.info(msg.format(page_num, cum_count))
            yield page

            # Update `last_id` for next page
            next_last_id = None
            while page:
                try:
                    next_last_id = page[-1].id
                except Exception:
                    logger.warning("Could not get ID from last result in page", exc_info=True)
                    page.pop()
                else:
                    break
            if next_last_id is not None:
                last_id = next_last_id
            else:
                raise RuntimeError("Could not find an object in the last page with an id")


def get_paginator(api, func_name: str, batch_size: Optional[int] = 1000):
    """Returns a Paginator object that returns pages of results from the given `func_name`.

    Example:

    .. code-block:: python

        import tator

        api = tator.get_api(host, token)
        media_paginator = tator.util.get_paginator(api, "get_media_list")
        page_iterator = media_paginator.paginate(project=project_id)
        all_media = []
        for page in page_iterator:
            for media in page:
                all_media.append(media)

    :param api: :class:`tator.TatorApi` object.
    :param str func_name:
    :param int batch_size:
    :raises: ValueError if the `func_name` does not match the expected pattern or if it does not
             exist.
    :returns: Paginator object
    """
    if not re.search(r"^get_.+_list(_by_id)?$", func_name):
        raise ValueError(
            f"Cannot paginate '{func_name}', expected a `get_<entity>_list` or "
            f"`get_<entity>_list_by_id` method"
        )
    try:
        fn = getattr(api, func_name)
    except AttributeError:
        similar = get_close_matches(func_name, dir(api), n=1)[0]
        raise ValueError(f"TatorApi has no function '{func_name}', did you mean '{similar}'?")
    return Paginator(fn, func_name, batch_size)



# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class ResolutionConfig(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'crf': 'int',
        'pixel_format': 'str',
        'preset': 'str',
        'resolution': 'int',
        'vcodec': 'str'
    }

    attribute_map = {
        'crf': 'crf',
        'pixel_format': 'pixel_format',
        'preset': 'preset',
        'resolution': 'resolution',
        'vcodec': 'vcodec'
    }

    def __init__(self, crf=23, pixel_format='yuv420p', preset='', resolution=None, vcodec='h264', local_vars_configuration=None):  # noqa: E501
        """ResolutionConfig - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._crf = None
        self._pixel_format = None
        self._preset = None
        self._resolution = None
        self._vcodec = None
        self.discriminator = None

        if crf is not None:
            self.crf = crf
        if pixel_format is not None:
            self.pixel_format = pixel_format
        if preset is not None:
            self.preset = preset
        if resolution is not None:
            self.resolution = resolution
        if vcodec is not None:
            self.vcodec = vcodec

    @property
    def crf(self):
        """
        Constant rate factor.

        :return: The crf of this ResolutionConfig. 
        :rtype: int
        """
        return self._crf

    @crf.setter
    def crf(self, crf):
        """
        Constant rate factor.

        :param crf: The crf of this ResolutionConfig.
        :type: int
        """
        if (self.local_vars_configuration.client_side_validation and
                crf is not None and crf > 63):  # noqa: E501
            raise ValueError("Invalid value for `crf`, must be a value less than or equal to `63`")  # noqa: E501
        if (self.local_vars_configuration.client_side_validation and
                crf is not None and crf < 0):  # noqa: E501
            raise ValueError("Invalid value for `crf`, must be a value greater than or equal to `0`")  # noqa: E501

        self._crf = crf

    @property
    def pixel_format(self):
        """
        Pixel format selection. Of note, compatibility varies significantly from codec to codec based on encoder support.SW Encoders: h264 when using libx264 supports:               yuv420p yuvj420p yuv422p yuvj422p yuv444p yuvj444p nv12 nv16 nv21 yuv420p10le yuv422p10le yuv444p10le nv20leh265 when using libsvt_hevc supports:           yuv420p yuvj420p yuv422p yuvj422p yuv444p yuvj444p nv12 nv16 nv21 yuv420p10le yuv422p10le yuv444p10le nv20leav1 when using libsvtav1 supports:              yuv420p          yuv422p          yuv444p                         yuv420p10le yuv422p10le yuv444p10leGenerally hardware encoders like QSV only support nv12(yuv420p) or p010le(yuv420p10le). Tator converts to the appropriate hardware format based on the specified pixel format.

        :return: The pixel_format of this ResolutionConfig. 
        :rtype: str
        """
        return self._pixel_format

    @pixel_format.setter
    def pixel_format(self, pixel_format):
        """
        Pixel format selection. Of note, compatibility varies significantly from codec to codec based on encoder support.SW Encoders: h264 when using libx264 supports:               yuv420p yuvj420p yuv422p yuvj422p yuv444p yuvj444p nv12 nv16 nv21 yuv420p10le yuv422p10le yuv444p10le nv20leh265 when using libsvt_hevc supports:           yuv420p yuvj420p yuv422p yuvj422p yuv444p yuvj444p nv12 nv16 nv21 yuv420p10le yuv422p10le yuv444p10le nv20leav1 when using libsvtav1 supports:              yuv420p          yuv422p          yuv444p                         yuv420p10le yuv422p10le yuv444p10leGenerally hardware encoders like QSV only support nv12(yuv420p) or p010le(yuv420p10le). Tator converts to the appropriate hardware format based on the specified pixel format.

        :param pixel_format: The pixel_format of this ResolutionConfig.
        :type: str
        """
        allowed_values = ["yuv420p", "yuvj420p", "yuv422p", "yuvj422p", "yuv444p", "yuvj444p", "yuv420p10le", "yuv422p10le", "yuv444p10le"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and pixel_format not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `pixel_format` ({0}), must be one of {1}"  # noqa: E501
                .format(pixel_format, allowed_values)
            )

        self._pixel_format = pixel_format

    @property
    def preset(self):
        """
        Codec Specific Preset (e.g. fast, medium, or 0,1,2)

        :return: The preset of this ResolutionConfig. 
        :rtype: str
        """
        return self._preset

    @preset.setter
    def preset(self, preset):
        """
        Codec Specific Preset (e.g. fast, medium, or 0,1,2)

        :param preset: The preset of this ResolutionConfig.
        :type: str
        """

        self._preset = preset

    @property
    def resolution(self):
        """
        Vertical Resolution

        :return: The resolution of this ResolutionConfig. 
        :rtype: int
        """
        return self._resolution

    @resolution.setter
    def resolution(self, resolution):
        """
        Vertical Resolution

        :param resolution: The resolution of this ResolutionConfig.
        :type: int
        """
        if (self.local_vars_configuration.client_side_validation and
                resolution is not None and resolution > 8192):  # noqa: E501
            raise ValueError("Invalid value for `resolution`, must be a value less than or equal to `8192`")  # noqa: E501
        if (self.local_vars_configuration.client_side_validation and
                resolution is not None and resolution < 0):  # noqa: E501
            raise ValueError("Invalid value for `resolution`, must be a value greater than or equal to `0`")  # noqa: E501

        self._resolution = resolution

    @property
    def vcodec(self):
        """
        Video codec.

        :return: The vcodec of this ResolutionConfig. 
        :rtype: str
        """
        return self._vcodec

    @vcodec.setter
    def vcodec(self, vcodec):
        """
        Video codec.

        :param vcodec: The vcodec of this ResolutionConfig.
        :type: str
        """
        allowed_values = ["h264", "av1"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and vcodec not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `vcodec` ({0}), must be one of {1}"  # noqa: E501
                .format(vcodec, allowed_values)
            )

        self._vcodec = vcodec

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResolutionConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResolutionConfig):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class LocalizationSpec(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'attributes': 'dict(str, object)',
        'elemental_id': 'str',
        'frame': 'int',
        'height': 'float',
        'mark': 'int',
        'media_id': 'int',
        'parent': 'float',
        'points': 'list[list[float]]',
        'type': 'int',
        'u': 'float',
        'user_elemental_id': 'str',
        'v': 'float',
        'version': 'int',
        'width': 'float',
        'x': 'float',
        'y': 'float'
    }

    attribute_map = {
        'attributes': 'attributes',
        'elemental_id': 'elemental_id',
        'frame': 'frame',
        'height': 'height',
        'mark': 'mark',
        'media_id': 'media_id',
        'parent': 'parent',
        'points': 'points',
        'type': 'type',
        'u': 'u',
        'user_elemental_id': 'user_elemental_id',
        'v': 'v',
        'version': 'version',
        'width': 'width',
        'x': 'x',
        'y': 'y'
    }

    def __init__(self, attributes=None, elemental_id=None, frame=None, height=None, mark=None, media_id=None, parent=None, points=None, type=None, u=None, user_elemental_id=None, v=None, version=None, width=None, x=None, y=None, local_vars_configuration=None):  # noqa: E501
        """LocalizationSpec - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._attributes = None
        self._elemental_id = None
        self._frame = None
        self._height = None
        self._mark = None
        self._media_id = None
        self._parent = None
        self._points = None
        self._type = None
        self._u = None
        self._user_elemental_id = None
        self._v = None
        self._version = None
        self._width = None
        self._x = None
        self._y = None
        self.discriminator = None

        if attributes is not None:
            self.attributes = attributes
        self.elemental_id = elemental_id
        self.frame = frame
        self.height = height
        if mark is not None:
            self.mark = mark
        self.media_id = media_id
        self.parent = parent
        self.points = points
        self.type = type
        self.u = u
        if user_elemental_id is not None:
            self.user_elemental_id = user_elemental_id
        self.v = v
        if version is not None:
            self.version = version
        self.width = width
        self.x = x
        self.y = y

    @property
    def attributes(self):
        """
        Object containing attribute values.

        :return: The attributes of this LocalizationSpec. 
        :rtype: dict(str, object)
        """
        return self._attributes

    @attributes.setter
    def attributes(self, attributes):
        """
        Object containing attribute values.

        :param attributes: The attributes of this LocalizationSpec.
        :type: dict(str, object)
        """

        self._attributes = attributes

    @property
    def elemental_id(self):
        """
        The elemental ID of the object.

        :return: The elemental_id of this LocalizationSpec. 
        :rtype: str
        """
        return self._elemental_id

    @elemental_id.setter
    def elemental_id(self, elemental_id):
        """
        The elemental ID of the object.

        :param elemental_id: The elemental_id of this LocalizationSpec.
        :type: str
        """

        self._elemental_id = elemental_id

    @property
    def frame(self):
        """
        Frame number of this localization if it is in a video.

        :return: The frame of this LocalizationSpec. 
        :rtype: int
        """
        return self._frame

    @frame.setter
    def frame(self, frame):
        """
        Frame number of this localization if it is in a video.

        :param frame: The frame of this LocalizationSpec.
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and frame is None:  # noqa: E501
            raise ValueError("Invalid value for `frame`, must not be `None`")  # noqa: E501

        self._frame = frame

    @property
    def height(self):
        """
        Normalized height of bounding box for `box` localization types.

        :return: The height of this LocalizationSpec. 
        :rtype: float
        """
        return self._height

    @height.setter
    def height(self, height):
        """
        Normalized height of bounding box for `box` localization types.

        :param height: The height of this LocalizationSpec.
        :type: float
        """
        if (self.local_vars_configuration.client_side_validation and
                height is not None and height > 1.0):  # noqa: E501
            raise ValueError("Invalid value for `height`, must be a value less than or equal to `1.0`")  # noqa: E501
        if (self.local_vars_configuration.client_side_validation and
                height is not None and height < 0.0):  # noqa: E501
            raise ValueError("Invalid value for `height`, must be a value greater than or equal to `0.0`")  # noqa: E501

        self._height = height

    @property
    def mark(self):
        """
        Revision number of this object on this version branch

        :return: The mark of this LocalizationSpec. 
        :rtype: int
        """
        return self._mark

    @mark.setter
    def mark(self, mark):
        """
        Revision number of this object on this version branch

        :param mark: The mark of this LocalizationSpec.
        :type: int
        """

        self._mark = mark

    @property
    def media_id(self):
        """
        Unique integer identifying a media.

        :return: The media_id of this LocalizationSpec. 
        :rtype: int
        """
        return self._media_id

    @media_id.setter
    def media_id(self, media_id):
        """
        Unique integer identifying a media.

        :param media_id: The media_id of this LocalizationSpec.
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and media_id is None:  # noqa: E501
            raise ValueError("Invalid value for `media_id`, must not be `None`")  # noqa: E501

        self._media_id = media_id

    @property
    def parent(self):
        """
        If a clone, the pk of the parent.

        :return: The parent of this LocalizationSpec. 
        :rtype: float
        """
        return self._parent

    @parent.setter
    def parent(self, parent):
        """
        If a clone, the pk of the parent.

        :param parent: The parent of this LocalizationSpec.
        :type: float
        """

        self._parent = parent

    @property
    def points(self):
        """
        List of normalized [x, y] pairs for `poly` localization types.

        :return: The points of this LocalizationSpec. 
        :rtype: list[list[float]]
        """
        return self._points

    @points.setter
    def points(self, points):
        """
        List of normalized [x, y] pairs for `poly` localization types.

        :param points: The points of this LocalizationSpec.
        :type: list[list[float]]
        """

        self._points = points

    @property
    def type(self):
        """
        Unique integer identifying a localization type.

        :return: The type of this LocalizationSpec. 
        :rtype: int
        """
        return self._type

    @type.setter
    def type(self, type):
        """
        Unique integer identifying a localization type.

        :param type: The type of this LocalizationSpec.
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and type is None:  # noqa: E501
            raise ValueError("Invalid value for `type`, must not be `None`")  # noqa: E501

        self._type = type

    @property
    def u(self):
        """
        Horizontal vector component for `line` localization types.

        :return: The u of this LocalizationSpec. 
        :rtype: float
        """
        return self._u

    @u.setter
    def u(self, u):
        """
        Horizontal vector component for `line` localization types.

        :param u: The u of this LocalizationSpec.
        :type: float
        """
        if (self.local_vars_configuration.client_side_validation and
                u is not None and u > 1.0):  # noqa: E501
            raise ValueError("Invalid value for `u`, must be a value less than or equal to `1.0`")  # noqa: E501
        if (self.local_vars_configuration.client_side_validation and
                u is not None and u < -1.0):  # noqa: E501
            raise ValueError("Invalid value for `u`, must be a value greater than or equal to `-1.0`")  # noqa: E501

        self._u = u

    @property
    def user_elemental_id(self):
        """
        Unique ID of the original user who created this. If permissions allow, will change the creating user to the one referenced by this elemental_id

        :return: The user_elemental_id of this LocalizationSpec. 
        :rtype: str
        """
        return self._user_elemental_id

    @user_elemental_id.setter
    def user_elemental_id(self, user_elemental_id):
        """
        Unique ID of the original user who created this. If permissions allow, will change the creating user to the one referenced by this elemental_id

        :param user_elemental_id: The user_elemental_id of this LocalizationSpec.
        :type: str
        """

        self._user_elemental_id = user_elemental_id

    @property
    def v(self):
        """
        Vertical vector component for `line` localization types.

        :return: The v of this LocalizationSpec. 
        :rtype: float
        """
        return self._v

    @v.setter
    def v(self, v):
        """
        Vertical vector component for `line` localization types.

        :param v: The v of this LocalizationSpec.
        :type: float
        """
        if (self.local_vars_configuration.client_side_validation and
                v is not None and v > 1.0):  # noqa: E501
            raise ValueError("Invalid value for `v`, must be a value less than or equal to `1.0`")  # noqa: E501
        if (self.local_vars_configuration.client_side_validation and
                v is not None and v < -1.0):  # noqa: E501
            raise ValueError("Invalid value for `v`, must be a value greater than or equal to `-1.0`")  # noqa: E501

        self._v = v

    @property
    def version(self):
        """
        Unique integer identifying the version.

        :return: The version of this LocalizationSpec. 
        :rtype: int
        """
        return self._version

    @version.setter
    def version(self, version):
        """
        Unique integer identifying the version.

        :param version: The version of this LocalizationSpec.
        :type: int
        """

        self._version = version

    @property
    def width(self):
        """
        Normalized width of bounding box for `box` localization types.

        :return: The width of this LocalizationSpec. 
        :rtype: float
        """
        return self._width

    @width.setter
    def width(self, width):
        """
        Normalized width of bounding box for `box` localization types.

        :param width: The width of this LocalizationSpec.
        :type: float
        """
        if (self.local_vars_configuration.client_side_validation and
                width is not None and width > 1.0):  # noqa: E501
            raise ValueError("Invalid value for `width`, must be a value less than or equal to `1.0`")  # noqa: E501
        if (self.local_vars_configuration.client_side_validation and
                width is not None and width < 0.0):  # noqa: E501
            raise ValueError("Invalid value for `width`, must be a value greater than or equal to `0.0`")  # noqa: E501

        self._width = width

    @property
    def x(self):
        """
        Normalized horizontal position of left edge of bounding box for `box` localization types, start of line for `line` localization types, or position of dot for `dot` localization types.

        :return: The x of this LocalizationSpec. 
        :rtype: float
        """
        return self._x

    @x.setter
    def x(self, x):
        """
        Normalized horizontal position of left edge of bounding box for `box` localization types, start of line for `line` localization types, or position of dot for `dot` localization types.

        :param x: The x of this LocalizationSpec.
        :type: float
        """
        if (self.local_vars_configuration.client_side_validation and
                x is not None and x > 1.0):  # noqa: E501
            raise ValueError("Invalid value for `x`, must be a value less than or equal to `1.0`")  # noqa: E501
        if (self.local_vars_configuration.client_side_validation and
                x is not None and x < 0.0):  # noqa: E501
            raise ValueError("Invalid value for `x`, must be a value greater than or equal to `0.0`")  # noqa: E501

        self._x = x

    @property
    def y(self):
        """
        Normalized vertical position of top edge of bounding box for `box` localization types, start of line for `line` localization types, or position of dot for `dot` localization types.

        :return: The y of this LocalizationSpec. 
        :rtype: float
        """
        return self._y

    @y.setter
    def y(self, y):
        """
        Normalized vertical position of top edge of bounding box for `box` localization types, start of line for `line` localization types, or position of dot for `dot` localization types.

        :param y: The y of this LocalizationSpec.
        :type: float
        """
        if (self.local_vars_configuration.client_side_validation and
                y is not None and y > 1.0):  # noqa: E501
            raise ValueError("Invalid value for `y`, must be a value less than or equal to `1.0`")  # noqa: E501
        if (self.local_vars_configuration.client_side_validation and
                y is not None and y < 0.0):  # noqa: E501
            raise ValueError("Invalid value for `y`, must be a value greater than or equal to `0.0`")  # noqa: E501

        self._y = y

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, LocalizationSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, LocalizationSpec):
            return True

        return self.to_dict() != other.to_dict()



from datetime import datetime, timezone

import tator


def test_date_before_now(host, token, project, video):
    tator_api = tator.get_api(host, token)
    video_obj = tator_api.get_media(video)
    assert datetime.now(timezone.utc) > video_obj.archive_status_date


def test_date_after_change(host, token, project, video):
    tator_api = tator.get_api(host, token)
    video_obj = tator_api.get_media(video)
    original_status_date = video_obj.archive_status_date

    tator_api.update_media(video, media_update={"archive_state": "to_archive"})
    video_obj = tator_api.get_media(video)
    to_archive_date = video_obj.archive_status_date
    assert datetime.now(timezone.utc) > to_archive_date > original_status_date

    tator_api.update_media(video, media_update={"archive_state": "to_live"})
    video_obj = tator_api.get_media(video)
    assert datetime.now(timezone.utc) > video_obj.archive_status_date > to_archive_date

def test_date_after_unrelated_change(host, token, project, video):
    tator_api = tator.get_api(host, token)
    video_obj = tator_api.get_media(video)
    original_status_date = video_obj.archive_status_date
    original_name = video_obj.name

    tator_api.update_media(video, media_update={"name": "new_name"})
    video_obj = tator_api.get_media(video)
    assert video_obj.archive_status_date == original_status_date

    tator_api.update_media(video, media_update={"name": original_name})
    video_obj = tator_api.get_media(video)
    assert video_obj.archive_status_date == original_status_date



# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class Favorite(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'entity_type_name': 'str',
        'id': 'int',
        'name': 'str',
        'page': 'int',
        'type': 'int',
        'user': 'int',
        'values': 'dict(str, object)'
    }

    attribute_map = {
        'entity_type_name': 'entity_type_name',
        'id': 'id',
        'name': 'name',
        'page': 'page',
        'type': 'type',
        'user': 'user',
        'values': 'values'
    }

    def __init__(self, entity_type_name=None, id=None, name=None, page=1, type=None, user=None, values=None, local_vars_configuration=None):  # noqa: E501
        """Favorite - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._entity_type_name = None
        self._id = None
        self._name = None
        self._page = None
        self._type = None
        self._user = None
        self._values = None
        self.discriminator = None

        if entity_type_name is not None:
            self.entity_type_name = entity_type_name
        if id is not None:
            self.id = id
        if name is not None:
            self.name = name
        if page is not None:
            self.page = page
        if type is not None:
            self.type = type
        if user is not None:
            self.user = user
        if values is not None:
            self.values = values

    @property
    def entity_type_name(self):
        """
        Name of entity type associated with the favorite

        :return: The entity_type_name of this Favorite. 
        :rtype: str
        """
        return self._entity_type_name

    @entity_type_name.setter
    def entity_type_name(self, entity_type_name):
        """
        Name of entity type associated with the favorite

        :param entity_type_name: The entity_type_name of this Favorite.
        :type: str
        """
        allowed_values = ["Localization", "State"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and entity_type_name not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `entity_type_name` ({0}), must be one of {1}"  # noqa: E501
                .format(entity_type_name, allowed_values)
            )

        self._entity_type_name = entity_type_name

    @property
    def id(self):
        """
        Unique integer identifying a favorite.

        :return: The id of this Favorite. 
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """
        Unique integer identifying a favorite.

        :param id: The id of this Favorite.
        :type: int
        """

        self._id = id

    @property
    def name(self):
        """
        Name of the favorite.

        :return: The name of this Favorite. 
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Name of the favorite.

        :param name: The name of this Favorite.
        :type: str
        """

        self._name = name

    @property
    def page(self):
        """
        Integer specifying page to display on. Should be 1-10.

        :return: The page of this Favorite. 
        :rtype: int
        """
        return self._page

    @page.setter
    def page(self, page):
        """
        Integer specifying page to display on. Should be 1-10.

        :param page: The page of this Favorite.
        :type: int
        """
        if (self.local_vars_configuration.client_side_validation and
                page is not None and page > 10):  # noqa: E501
            raise ValueError("Invalid value for `page`, must be a value less than or equal to `10`")  # noqa: E501
        if (self.local_vars_configuration.client_side_validation and
                page is not None and page < 1):  # noqa: E501
            raise ValueError("Invalid value for `page`, must be a value greater than or equal to `1`")  # noqa: E501

        self._page = page

    @property
    def type(self):
        """
        Unique integer identifying entity type of this entry.

        :return: The type of this Favorite. 
        :rtype: int
        """
        return self._type

    @type.setter
    def type(self, type):
        """
        Unique integer identifying entity type of this entry.

        :param type: The type of this Favorite.
        :type: int
        """

        self._type = type

    @property
    def user(self):
        """
        Unique integer identifying a user.

        :return: The user of this Favorite. 
        :rtype: int
        """
        return self._user

    @user.setter
    def user(self, user):
        """
        Unique integer identifying a user.

        :param user: The user of this Favorite.
        :type: int
        """
        if (self.local_vars_configuration.client_side_validation and
                user is not None and user < 1):  # noqa: E501
            raise ValueError("Invalid value for `user`, must be a value greater than or equal to `1`")  # noqa: E501

        self._user = user

    @property
    def values(self):
        """
        Attribute name/value pairs.

        :return: The values of this Favorite. 
        :rtype: dict(str, object)
        """
        return self._values

    @values.setter
    def values(self, values):
        """
        Attribute name/value pairs.

        :param values: The values of this Favorite.
        :type: dict(str, object)
        """

        self._values = values

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Favorite):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Favorite):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class MediaSpec(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'attributes': 'dict(str, object)',
        'codec': 'str',
        'elemental_id': 'str',
        'fps': 'float',
        'gid': 'str',
        'height': 'int',
        'md5': 'str',
        'name': 'str',
        'num_frames': 'int',
        'reference_only': 'int',
        'section': 'str',
        'section_id': 'int',
        'summary_level': 'int',
        'thumbnail_gif_url': 'str',
        'thumbnail_url': 'str',
        'type': 'int',
        'uid': 'str',
        'url': 'str',
        'user_elemental_id': 'str',
        'width': 'int'
    }

    attribute_map = {
        'attributes': 'attributes',
        'codec': 'codec',
        'elemental_id': 'elemental_id',
        'fps': 'fps',
        'gid': 'gid',
        'height': 'height',
        'md5': 'md5',
        'name': 'name',
        'num_frames': 'num_frames',
        'reference_only': 'reference_only',
        'section': 'section',
        'section_id': 'section_id',
        'summary_level': 'summary_level',
        'thumbnail_gif_url': 'thumbnail_gif_url',
        'thumbnail_url': 'thumbnail_url',
        'type': 'type',
        'uid': 'uid',
        'url': 'url',
        'user_elemental_id': 'user_elemental_id',
        'width': 'width'
    }

    def __init__(self, attributes=None, codec=None, elemental_id=None, fps=None, gid=None, height=None, md5=None, name=None, num_frames=None, reference_only=0, section=None, section_id=None, summary_level=None, thumbnail_gif_url=None, thumbnail_url=None, type=None, uid=None, url=None, user_elemental_id=None, width=None, local_vars_configuration=None):  # noqa: E501
        """MediaSpec - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._attributes = None
        self._codec = None
        self._elemental_id = None
        self._fps = None
        self._gid = None
        self._height = None
        self._md5 = None
        self._name = None
        self._num_frames = None
        self._reference_only = None
        self._section = None
        self._section_id = None
        self._summary_level = None
        self._thumbnail_gif_url = None
        self._thumbnail_url = None
        self._type = None
        self._uid = None
        self._url = None
        self._user_elemental_id = None
        self._width = None
        self.discriminator = None

        self.attributes = attributes
        self.codec = codec
        if elemental_id is not None:
            self.elemental_id = elemental_id
        self.fps = fps
        if gid is not None:
            self.gid = gid
        self.height = height
        self.md5 = md5
        self.name = name
        self.num_frames = num_frames
        if reference_only is not None:
            self.reference_only = reference_only
        if section is not None:
            self.section = section
        if section_id is not None:
            self.section_id = section_id
        if summary_level is not None:
            self.summary_level = summary_level
        if thumbnail_gif_url is not None:
            self.thumbnail_gif_url = thumbnail_gif_url
        if thumbnail_url is not None:
            self.thumbnail_url = thumbnail_url
        self.type = type
        if uid is not None:
            self.uid = uid
        if url is not None:
            self.url = url
        if user_elemental_id is not None:
            self.user_elemental_id = user_elemental_id
        self.width = width

    @property
    def attributes(self):
        """
        Object containing attribute values.

        :return: The attributes of this MediaSpec. 
        :rtype: dict(str, object)
        """
        return self._attributes

    @attributes.setter
    def attributes(self, attributes):
        """
        Object containing attribute values.

        :param attributes: The attributes of this MediaSpec.
        :type: dict(str, object)
        """

        self._attributes = attributes

    @property
    def codec(self):
        """
        Codec for videos.

        :return: The codec of this MediaSpec. 
        :rtype: str
        """
        return self._codec

    @codec.setter
    def codec(self, codec):
        """
        Codec for videos.

        :param codec: The codec of this MediaSpec.
        :type: str
        """

        self._codec = codec

    @property
    def elemental_id(self):
        """
        Unique ID of an element

        :return: The elemental_id of this MediaSpec. 
        :rtype: str
        """
        return self._elemental_id

    @elemental_id.setter
    def elemental_id(self, elemental_id):
        """
        Unique ID of an element

        :param elemental_id: The elemental_id of this MediaSpec.
        :type: str
        """

        self._elemental_id = elemental_id

    @property
    def fps(self):
        """
        Frame rate for videos.

        :return: The fps of this MediaSpec. 
        :rtype: float
        """
        return self._fps

    @fps.setter
    def fps(self, fps):
        """
        Frame rate for videos.

        :param fps: The fps of this MediaSpec.
        :type: float
        """

        self._fps = fps

    @property
    def gid(self):
        """
        Group ID for the upload group of this media.

        :return: The gid of this MediaSpec. 
        :rtype: str
        """
        return self._gid

    @gid.setter
    def gid(self, gid):
        """
        Group ID for the upload group of this media.

        :param gid: The gid of this MediaSpec.
        :type: str
        """

        self._gid = gid

    @property
    def height(self):
        """
        Vertical resolution in pixels.

        :return: The height of this MediaSpec. 
        :rtype: int
        """
        return self._height

    @height.setter
    def height(self, height):
        """
        Vertical resolution in pixels.

        :param height: The height of this MediaSpec.
        :type: int
        """

        self._height = height

    @property
    def md5(self):
        """
        MD5 sum of the media file.

        :return: The md5 of this MediaSpec. 
        :rtype: str
        """
        return self._md5

    @md5.setter
    def md5(self, md5):
        """
        MD5 sum of the media file.

        :param md5: The md5 of this MediaSpec.
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and md5 is None:  # noqa: E501
            raise ValueError("Invalid value for `md5`, must not be `None`")  # noqa: E501

        self._md5 = md5

    @property
    def name(self):
        """
        Name of the file.

        :return: The name of this MediaSpec. 
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Name of the file.

        :param name: The name of this MediaSpec.
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def num_frames(self):
        """
        Number of frames for videos.

        :return: The num_frames of this MediaSpec. 
        :rtype: int
        """
        return self._num_frames

    @num_frames.setter
    def num_frames(self, num_frames):
        """
        Number of frames for videos.

        :param num_frames: The num_frames of this MediaSpec.
        :type: int
        """

        self._num_frames = num_frames

    @property
    def reference_only(self):
        """
        Do not import the media resources into Tator. This causes the image to be accessed at the supplied URL upon access in the UI. Attempts loading the image from the URL, but fall back to other records if the URL is not accessible.

        :return: The reference_only of this MediaSpec. 
        :rtype: int
        """
        return self._reference_only

    @reference_only.setter
    def reference_only(self, reference_only):
        """
        Do not import the media resources into Tator. This causes the image to be accessed at the supplied URL upon access in the UI. Attempts loading the image from the URL, but fall back to other records if the URL is not accessible.

        :param reference_only: The reference_only of this MediaSpec.
        :type: int
        """
        if (self.local_vars_configuration.client_side_validation and
                reference_only is not None and reference_only > 1):  # noqa: E501
            raise ValueError("Invalid value for `reference_only`, must be a value less than or equal to `1`")  # noqa: E501

        self._reference_only = reference_only

    @property
    def section(self):
        """
        Media section name.

        :return: The section of this MediaSpec. 
        :rtype: str
        """
        return self._section

    @section.setter
    def section(self, section):
        """
        Media section name.

        :param section: The section of this MediaSpec.
        :type: str
        """

        self._section = section

    @property
    def section_id(self):
        """
        Media section ID. If given `section` is ignored.

        :return: The section_id of this MediaSpec. 
        :rtype: int
        """
        return self._section_id

    @section_id.setter
    def section_id(self, section_id):
        """
        Media section ID. If given `section` is ignored.

        :param section_id: The section_id of this MediaSpec.
        :type: int
        """

        self._section_id = section_id

    @property
    def summary_level(self):
        """
        If supplied, this video is best summarized at this frame interval

        :return: The summary_level of this MediaSpec. 
        :rtype: int
        """
        return self._summary_level

    @summary_level.setter
    def summary_level(self, summary_level):
        """
        If supplied, this video is best summarized at this frame interval

        :param summary_level: The summary_level of this MediaSpec.
        :type: int
        """

        self._summary_level = summary_level

    @property
    def thumbnail_gif_url(self):
        """
        Upload URL for the video gif thumbnail if already generated.

        :return: The thumbnail_gif_url of this MediaSpec. 
        :rtype: str
        """
        return self._thumbnail_gif_url

    @thumbnail_gif_url.setter
    def thumbnail_gif_url(self, thumbnail_gif_url):
        """
        Upload URL for the video gif thumbnail if already generated.

        :param thumbnail_gif_url: The thumbnail_gif_url of this MediaSpec.
        :type: str
        """

        self._thumbnail_gif_url = thumbnail_gif_url

    @property
    def thumbnail_url(self):
        """
        Upload URL for the media thumbnail if already generated.

        :return: The thumbnail_url of this MediaSpec. 
        :rtype: str
        """
        return self._thumbnail_url

    @thumbnail_url.setter
    def thumbnail_url(self, thumbnail_url):
        """
        Upload URL for the media thumbnail if already generated.

        :param thumbnail_url: The thumbnail_url of this MediaSpec.
        :type: str
        """

        self._thumbnail_url = thumbnail_url

    @property
    def type(self):
        """
        Unique integer identifying a media type. Use -1 to automatically select the media type if only one media type exists in a project.

        :return: The type of this MediaSpec. 
        :rtype: int
        """
        return self._type

    @type.setter
    def type(self, type):
        """
        Unique integer identifying a media type. Use -1 to automatically select the media type if only one media type exists in a project.

        :param type: The type of this MediaSpec.
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and type is None:  # noqa: E501
            raise ValueError("Invalid value for `type`, must not be `None`")  # noqa: E501
        if (self.local_vars_configuration.client_side_validation and
                type is not None and type < -1):  # noqa: E501
            raise ValueError("Invalid value for `type`, must be a value greater than or equal to `-1`")  # noqa: E501

        self._type = type

    @property
    def uid(self):
        """
        Unique ID for the upload of this media.

        :return: The uid of this MediaSpec. 
        :rtype: str
        """
        return self._uid

    @uid.setter
    def uid(self, uid):
        """
        Unique ID for the upload of this media.

        :param uid: The uid of this MediaSpec.
        :type: str
        """

        self._uid = uid

    @property
    def url(self):
        """
        Upload URL for the image if this is an image type, URL of hosted original media if this is a video type. For video types this field is just for reference.

        :return: The url of this MediaSpec. 
        :rtype: str
        """
        return self._url

    @url.setter
    def url(self, url):
        """
        Upload URL for the image if this is an image type, URL of hosted original media if this is a video type. For video types this field is just for reference.

        :param url: The url of this MediaSpec.
        :type: str
        """

        self._url = url

    @property
    def user_elemental_id(self):
        """
        Unique ID of the original user who created this. If permissions allow, will change the creating user to the one referenced by this elemental_id

        :return: The user_elemental_id of this MediaSpec. 
        :rtype: str
        """
        return self._user_elemental_id

    @user_elemental_id.setter
    def user_elemental_id(self, user_elemental_id):
        """
        Unique ID of the original user who created this. If permissions allow, will change the creating user to the one referenced by this elemental_id

        :param user_elemental_id: The user_elemental_id of this MediaSpec.
        :type: str
        """

        self._user_elemental_id = user_elemental_id

    @property
    def width(self):
        """
        Horizontal resolution in pixels.

        :return: The width of this MediaSpec. 
        :rtype: int
        """
        return self._width

    @width.setter
    def width(self, width):
        """
        Horizontal resolution in pixels.

        :param width: The width of this MediaSpec.
        :type: int
        """

        self._width = width

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, MediaSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, MediaSpec):
            return True

        return self.to_dict() != other.to_dict()




import logging
import os
import requests
import tempfile

import pytest
import tator
from tator.util._upload_file import _upload_file
import json
import base64

logger = logging.getLogger(__name__)

def _missing_upload_file(
        host: str,
        token: str,
        project: int) -> None:
    """ Test missing upload file for SaveGenericFile endpoint

    :param host: Project URL
    :param token: User token used for connecting to the host
    :param project: Unique identifier of test project
    """

    tator_api = tator.get_api(host=host, token=token)
    spec = tator.models.GenericFileSpec(name="test.file", upload_url="not_there")
    with pytest.raises(tator.openapi.tator_openapi.exceptions.ApiException):
        tator_api.save_generic_file(project=project, generic_file_spec=spec)

def _upload_generic_file(
        host: str,
        token: str,
        project: int,
        file_id: int,
        file_path: str="test.file"):
    """ Uploads the provided manifest file

    :param host: Project URL
    :param token: User token used for connecting to the host
    :param project: Unique identifier of test project
    :param file_id: Unique identifier of associated file
    :param file_path: Local file to be uploaded and saved
    :returns: Associated key of uploaded file
    """

    # Setup the interface to tator
    tator_api = tator.get_api(host=host, token=token)

    # Create the temporary file
    fd, local_file = tempfile.mkstemp()
    try:
        with os.fdopen(fd, 'w') as file_handle:
            file_handle.write("Yo!")

        # Upload the file
        logger.info(f"Created temporary file: {local_file}")
        for progress, upload_info in _upload_file(api=tator_api, project=project, path=local_file, file_id=file_id):
            logger.info(f"Upload progress: {progress}%")

    finally:
        os.remove(local_file)

    return upload_info.key

def _make_all_attribute_types() -> list:
    """ Taken from conftest.py
    :returns: One of each possible attribute dtype with generic values
    """
    return [
        dict(
            name='test_bool',
            dtype='bool',
            default=True,
        ),
        dict(
            name='test_int',
            dtype='int',
            default=0,
            minimum=-1000,
            maximum=1000,
        ),
        dict(
            name='test_float',
            dtype='float',
            default=0.0,
            minimum=-1000.0,
            maximum=1000.0,
        ),
        dict(
            name='test_enum',
            dtype='enum',
            choices=['a', 'b', 'c'],
            default='a',
        ),
        dict(
            name='test_string',
            dtype='string',
            default='asdf',
        ),
        dict(
            name='test_datetime',
            dtype='datetime',
            use_current=True,
        ),
        dict(
            name='test_geopos',
            dtype='geopos',
            default=[-179.0, -89.0],
        ),
        dict(
            name='test_float_array',
            dtype='float_array',
            default=[0.0, 0.0, 0.0],
            size=3,
        )
    ]

def test_file_type_crud(
        host: str,
        token: str,
        project: int) -> None:
    """ Test the get, post, patch, and delete FileType endpoints

    :param host: Project URL
    :param token: User token used for connecting to the host
    :param project: Unique identifier of test project

    """

    # Setup the interface to Tator
    tator_api = tator.get_api(host=host, token=token)

    # Test creating FileType objects with all available fields
    response = tator_api.create_file_type(project=project, file_type_spec={
        "name": "file_type_with_attrs",
        "description": "Test file type with attributes",
        "project": project,
        "attribute_types": _make_all_attribute_types()
    })
    file_type_id_1 = response.id
    assert file_type_id_1 >= 0

    # Test creating FileType object without attribute and description
    response = tator_api.create_file_type(project=project, file_type_spec={
        "name": "file_type_without_attrs",
        "project": project,
    })
    file_type_id_2 = response.id
    assert file_type_id_2 >= 0

    # Test retrieval by ID and list
    files_types = tator_api.get_file_type_list(project=project)
    found_file_type_id_1 = False
    found_file_type_id_2 = False
    for file_type in files_types:
        if file_type.id == file_type_id_1:
            found_file_type_id_1 = True
        elif file_type.id == file_type_id_2:
            found_file_type_id_2 = True

    assert found_file_type_id_1
    assert found_file_type_id_2

    file_type = tator_api.get_file_type(id=file_type_id_1)
    assert file_type.id == file_type_id_1

    file_type = tator_api.get_file_type(id=file_type_id_2)
    assert file_type.id == file_type_id_2

    # Update the FileType object
    tator_api.update_file_type(id=file_type_id_2, file_type_update={
        "name": "Another name",
        "description": "Another description"
    })
    file_type = tator_api.get_file_type(id=file_type_id_2)
    assert file_type.name == "Another name"
    assert file_type.description == "Another description"

    # Test deleting
    tator_api.delete_file_type(id=file_type_id_2)

    with pytest.raises(tator.openapi.tator_openapi.exceptions.ApiException):
        file_type = tator_api.get_file_type(id=file_type_id_2)

def test_file_crud(
        host: str,
        token: str,
        project: int) -> None:
    """ Test the get, post, patch, and delete File endpoints. Also test attribute type updates.

    :param host: Project URL
    :param token: User token used for connecting to the host
    :param project: Unique identifier of test project

    FileTypeA
    LabelA - string dtype
    LabelB - int dtype

    FileTypeB
    LabelA - string dtype

    This will have 10 different File objects with FileType - A.
    This will have 100 different File objects with FileType - B.

    """

    # Setup the interface to Tator
    tator_api = tator.get_api(host=host, token=token)

    # Create two FileTypes so we can test searching
    response = tator_api.create_file_type(project=project, file_type_spec={
        "name": "file_type_without_attrs",
        "project": project,
        "attribute_types": [
            dict(
                name="LabelA",
                dtype="int",
                default=0,
                minimum=-1000,
                maximum=1000,
            ),
            dict(
                name="LabelB",
                dtype="string",
                default="",
            ),
        ]
    })
    file_type_a_id = response.id

    response = tator_api.create_file_type(project=project, file_type_spec={
        "name": "file_type_without_attrs",
        "project": project,
        "attribute_types": [
            dict(
                name="LabelA",
                dtype="int",
                default=0,
                minimum=-1000,
                maximum=1000,
            ),
            dict(
                name="LabelB",
                dtype="string",
                default="",
            ),
        ]
    })
    file_type_b_id = response.id

    # Create the files of each type
    file_a_ids = []
    for idx in range(10):
        response = tator_api.create_file(project=project, file_spec=dict(
            name=f"File_A_{idx}",
            description="hey",
            type=file_type_a_id,
            attributes=dict(
                LabelA=idx,
                LabelB="seeya"
            )
        ))
        key = _upload_generic_file(
            host=host, token=token, project=project, file_id=response.id, file_path="test.data")
        tator_api.update_file(id=response.id, file_update={"path": key})
        file_a_ids.append({
            "file_id": response.id,
            "key": key
        })

    file_b_ids = []
    for idx in range(50):
        response = tator_api.create_file(project=project, file_spec=dict(
            name=f"File_B_{idx}",
            description="hey",
            type=file_type_b_id,
            attributes=dict(
                LabelB="bye"
            )
        ))
        key = _upload_generic_file(
            host=host, token=token, project=project, file_id=response.id, file_path="test.data")
        tator_api.update_file(id=response.id, file_update={"path": key})
        file_b_ids.append({
            "file_id": response.id,
            "key": key
        })

    # Get all the files
    file_list = tator_api.get_file_list(project=project)
    file_a_count = 0
    file_b_count = 0
    for file_obj in file_list:
        for file_info in file_a_ids:
            if file_obj.id == file_info["file_id"]:
                file_a_count += 1
                break

        for file_info in file_b_ids:
            if file_obj.id == file_info["file_id"]:
                file_b_count += 1
                break

    assert file_a_count == len(file_a_ids)
    assert file_b_count == len(file_b_ids)

    # Grab only the data associated with the FileTypes
    files = tator_api.get_file_list(project=project, type=file_type_a_id)
    assert len(files) == len(file_a_ids)

    files = tator_api.get_file_list(project=project, type=file_type_b_id)
    assert len(files) == len(file_b_ids)

    files = tator_api.get_file_list(project=project, attribute=["LabelB::seeya"])
    assert len(files) == len(file_a_ids)

    blob=base64.b64encode(json.dumps({'attribute': 'LabelB', 'operation': 'eq', 'value': 'seeya'}).encode('ascii'))
    files = tator_api.get_file_list(project=project, encoded_search=blob)
    assert len(files) == len(file_a_ids)

    # Update a bunch of the entries and verify the updates are valid and
    # the search still works correctly
    for info in file_a_ids:
        current_id = info["file_id"]
        tator_api.update_file(id=current_id, file_update={
            "name": "new_name",
            "description": "new_description",
            "attributes": {"LabelA": -100}
        })
    files = tator_api.get_file_list(project=project, type=file_type_a_id)
    for file_obj in files:
        update_check = \
            file_obj.name == "new_name" and \
            file_obj.description == "new_description" and \
            file_obj.attributes["LabelA"] == -100 and \
            file_obj.attributes["LabelB"] == "seeya"
        if not update_check:
            print(file_obj)
            break

    assert update_check

    blob=base64.b64encode(json.dumps({'attribute': 'LabelA', 'operation': 'lt', 'value': 0}).encode('ascii'))
    files = tator_api.get_file_list(project=project, encoded_search=blob)
    assert len(files) == len(file_a_ids)

    # Check to see if pagination works
    page_size = 10
    num_pages = (len(file_b_ids) / page_size) - 1
    start = 0
    stop = page_size
    page = 0
    all_files = set()
    files = tator_api.get_file_list(project=project, type=file_type_b_id, start=start, stop=stop)
    for file in files:
        all_files.add(file.id)
    files = tator_api.get_file_list(project=project, type=file_type_b_id, start=start+(page*page_size), stop=stop+(page*page_size))
    while files:
        for file in files:
            all_files.add(file.id)
        page += 1
        files = tator_api.get_file_list(project=project, type=file_type_b_id, start=start+(page*page_size), stop=stop+(page*page_size))

    files = tator_api.get_file_list(project=project, type=file_type_b_id)
    blah = [file.id for file in files]
    assert len(all_files) == len(file_b_ids)

    # Delete 2 entries from each file group. Verify they were deleted.
    tator_api.delete_file(id=file_a_ids[0]["file_id"])
    tator_api.delete_file(id=file_a_ids[-1]["file_id"])
    tator_api.delete_file(id=file_b_ids[0]["file_id"])
    tator_api.delete_file(id=file_b_ids[-1]["file_id"])
    file_list = tator_api.get_file_list(project=project)
    file_a_count = 0
    file_b_count = 0
    for file_obj in file_list:
        for file_info in file_a_ids:
            if file_obj.id == file_info["file_id"]:
                file_a_count += 1
                break

        for file_info in file_b_ids:
            if file_obj.id == file_info["file_id"]:
                file_b_count += 1
                break

    assert file_a_count == len(file_a_ids) - 2
    assert file_b_count == len(file_b_ids) - 2


def test_upload_generic_file(host, token, project):
    TEST_STR = b"foo"
    # Setup the interface to Tator
    tator_api = tator.get_api(host=host, token=token)

    # Create a FileType
    response = tator_api.create_file_type(
        project=project,
        file_type_spec={
            "name": "generic_upload_file_type",
            "project": project,
            "attribute_types": [
                {"name": "test", "dtype": "int", "default": 0, "minimum": -1000, "maximum": 1000}
            ],
        },
    )
    type_id = response.id

    # Create a file locally and test uploading/downloading
    with tempfile.NamedTemporaryFile() as temp_file:
        path = temp_file.name
        fname = os.path.basename(path)

        temp_file.write(TEST_STR)
        temp_file.flush()

        for progress, response in tator.util.upload_generic_file(tator_api, type_id, path, "File description", fname):
            print(f"Progress: {progress}%")
    assert response.name == fname
    download_info = tator_api.get_download_info(project=project, download_info_spec={"keys": [response.path]})

    assert len(download_info) == 1
    assert download_info[0].key == response.path
    response = requests.get(download_info[0].url, stream=True)
    assert response.content == TEST_STR



import tator


def test_string_attributes_bulk_update(host, token, project, video_type, video):
    tator_api = tator.get_api(host, token)

    # Put a newline character in a string attribute
    media_bulk_update = {"attributes": {"test_string": "test\nnewline"}, "ids": [video]}
    tator_api.update_media_list(project, media_bulk_update)
    video_obj = tator_api.get_media(video)
    assert video_obj.attributes["test_string"] == media_bulk_update["attributes"]["test_string"]

    # Put a backslash character in a string attribute
    media_bulk_update["attributes"]["test_string"] = "test\\backslash"
    tator_api.update_media_list(project, media_bulk_update)
    video_obj = tator_api.get_media(video)
    assert video_obj.attributes["test_string"] == media_bulk_update["attributes"]["test_string"]

    # Put a double quote character in a string attribute
    media_bulk_update["attributes"]["test_string"] = 'test "double" quotes'
    tator_api.update_media_list(project, media_bulk_update)
    video_obj = tator_api.get_media(video)
    assert video_obj.attributes["test_string"] == media_bulk_update["attributes"]["test_string"]




from pprint import pformat
import datetime
import random
from time import sleep
import uuid
from collections import Counter

import tator
from ._common import assert_close_enough


def random_localization(project, poly_type, video_obj, post=False):
    attributes = {
        "test_bool": random.choice([False, True]),
        "test_int": random.randint(-1000, 1000),
        "test_float": random.uniform(-1000.0, 1000.0),
        "test_enum": random.choice(["a", "b", "c"]),
        "test_string": str(uuid.uuid1()),
        "test_datetime": datetime.datetime.now().isoformat(),
        "test_geopos": [random.uniform(-180.0, 180.0), random.uniform(-90.0, 90.0)],
        "test_float_array": [random.uniform(-1.0, 1.0) for _ in range(3)],
    }
    out = {
        "points": [
            [random.uniform(0.0, 1.0), random.uniform(0.0, 1.0)]
            for _ in range(random.randint(1, 10))
        ],
        "project": project,
        "type": poly_type,
        "media_id": video_obj.id,
        "frame": random.randint(0, video_obj.num_frames - 1),
        "attributes": attributes,
    }

    return {**out}


def test_poly(host, token, project, video_type, video, poly_type):
    tator_api = tator.get_api(host, token)
    video_obj = tator_api.get_media(video)

    # These fields will not be checked for object equivalence after patch.
    exclude = ["project", "type", "media_id", "id", "type", "user", "ids", "in_place"]

    # Test single create.
    poly = random_localization(project, poly_type, video_obj, post=True)
    response = tator_api.create_localization_list(project, poly)
    assert isinstance(response, tator.models.CreateListResponse)
    poly_id = response.id[0]
    response = tator_api.get_localization(poly_id)

    # Patch single poly.
    patch = random_localization(project, poly_type, video_obj)
    patch_obj = {**patch, "in_place": 1}
    response = tator_api.update_localization(poly_id, localization_update=patch_obj)
    assert "message" in response.to_dict()
    print(response.message)

    # Get single poly.
    updated_poly = tator_api.get_localization(poly_id)
    assert_close_enough(patch, updated_poly, exclude)

    # Get poly by ID.
    poly_by_id = tator_api.get_localization_list_by_id(project, {"ids": [poly_id]})
    assert len(poly_by_id) == 1
    poly_by_id = poly_by_id[0]
    assert_close_enough(updated_poly, poly_by_id, exclude)

    # Delete single poly.
    response = tator_api.delete_localization(poly_id)
    assert "message" in response.to_dict()
    print(response.message)



import os
import tempfile
import logging
from uuid import uuid1

from PIL import Image, ImageSequence

from ._upload_file import _upload_file
from ._download_file import _download_file
from ..openapi.tator_openapi.models import MessageResponse

logger = logging.getLogger(__name__)
GIF_SZ = 256

PLACEHOLDER_DIR = os.path.dirname(os.path.abspath(__file__))
PLACEHOLDER_THUMB_PATH = os.path.join(PLACEHOLDER_DIR, "tator-symbol.png")
PLACEHOLDER_THUMB_GIF_PATH = os.path.join(PLACEHOLDER_DIR, "tator-symbol.gif")


def _merge_gifs(layout, gif_list, output_filename) -> bool:
    # Calculate the dimensions of the merged GIF
    gifs = []
    try:
        # Open all
        frame_iters = []
        for gif_path in gif_list:
            gifs.append(Image.open(gif_path))
            frame_iters.append(ImageSequence.Iterator(gifs[-1]))

        merged_frames = []
        rows, columns = layout
        merged_height = 256 * rows
        merged_width = 256 * columns
        for single_frames in zip(*frame_iters):
            merged_frames.append(Image.new("RGB", (merged_width, merged_height)))
            for idx, single_frame in enumerate(single_frames):
                y_offset = GIF_SZ * (idx // rows)
                x_offset = 2 * GIF_SZ * (idx % rows)
                merged_frames[-1].paste(
                    single_frame.resize((2 * GIF_SZ, GIF_SZ), Image.LANCZOS), (x_offset, y_offset)
                )
        # Save the merged GIF
        try:
            merged_frames[0].save(
                output_filename,
                save_all=True,
                append_images=merged_frames[1:],
            )
        except Exception:
            logger.warning(
                "Could not save merged gif '%s', using placeholder", output_filename, exc_info=True
            )
            return False
    finally:
        for gif in gifs:
            gif.close()
    return True


def make_multi_stream(
    api, type_id, layout, name, media_ids, section=None, quality=None, frame_offset=None, section_id=None
):
    """Creates a multiview from the given arguments

    :param api: The Tator client
    :type api: tator.openapi.TatorApi
    :param type_id: Unique integer identifying a multi-stream media type.
    :type type_id: int
    :param layout: 2 element list of integers defining layout as [rows, cols].
    :type layout: List[int]
    :param name: Name of the file to use
    :type name: str
    :param media_ids: List of media_ids to multi-stream
    :type media_ids: List[int]
    :param section: [Optional] Section name; if it does not exist, it will be created. Either section or section_id must be provided.
    :type section: str
    :param quality: [Optional] The desired resolution to pull from the single media
    :type quality: List[int]
    :param frame_offset: [Optional] Frame offsets to apply to each stream.
    :type frame_offset: List[int]
    :param section_id: [Optional] Section ID to use instead of creating a new section.
    :type section_id: int
    :returns: Response from media object creation.
    """
    tiled_thumb_filename = "tiled_thumb.jpg"
    tiled_gif_filename = "tiled_gif.gif"
    thumb_path = PLACEHOLDER_THUMB_PATH
    thumb_gif_path = PLACEHOLDER_THUMB_GIF_PATH

    # use a multi extension
    name += ".multi"

    # Fetch the media type
    multi_stream_type = api.get_media_type(type_id)
    project = multi_stream_type.project

    # Fetch the section. If it does not exist, create it.
    if section_id:
        section_obj = api.get_section(section_id)
    elif section:
        sections = api.get_section_list(project, name=section)
        if len(sections) == 0:
            section_spec = {"name": section, "tator_user_sections": str(uuid1())}
            response = api.create_section(project, section_spec=section_spec)
            section_obj = api.get_section_list(project, name=section)[0]
        else:
            section_obj = sections[0]
    else:
        raise RuntimeError("Either section or section_id must be provided")

    n_ids = len(media_ids)
    n_spots = layout[0] * layout[1]
    if n_ids != n_spots:
        raise RuntimeError(
            f"{layout[0]} by {layout[1]} layout requires {n_spots} videos, received {n_ids}"
        )

    media_objects = api.get_media_list(project, media_id=media_ids)
    n_objs = len(media_objects)
    if n_objs != n_ids:
        raise RuntimeError(f"Requested {n_ids} media objects, but only found {n_objs}")
    media_lookup = {media.id: media for media in media_objects}

    attributes = {}
    if section:
        attributes["tator_user_sections"] = section_obj.tator_user_sections

    # Download the thumbnails into a temporary directory
    with tempfile.TemporaryDirectory() as d:
        input_files = []
        md5 = ""
        for pos, media_id in enumerate(media_ids):
            media = media_lookup[media_id]
            md5 = media.md5
            if media.media_files is None:
                continue
            thumbnail_gifs = media.media_files.thumbnail_gif
            if thumbnail_gifs:
                thumb_gif = thumbnail_gifs[0].path
                thumb_name = os.path.join(d, f"gif_{pos:09d}.gif")
                try:
                    for _ in _download_file(api, media.project, thumb_gif, thumb_name):
                        pass
                except Exception:
                    logger.warning(
                        "Could not download thumbnail gif from %d", media_id, exc_info=True
                    )
                else:
                    input_files.append(thumb_name)

        new_thumb_gif_path = os.path.join(d, tiled_gif_filename)
        if len(input_files) == n_ids and _merge_gifs(
            layout=layout, gif_list=input_files, output_filename=new_thumb_gif_path
        ):
            thumb_gif_path = new_thumb_gif_path
            try:
                new_thumb_path = os.path.join(d, tiled_thumb_filename)
                with Image.open(thumb_gif_path) as img:
                    first_frame = img.convert("RGB")
                    first_frame.save(new_thumb_path)
            except Exception:
                logger.warning("Could not create thumbnail image from gif", exc_info=True)
            else:
                thumb_path = new_thumb_path

        media_spec = {
            "attributes": attributes,
            "name": name,
            "md5": md5,
            "section_id": section_obj.id,
            "type": type_id,
        }
        try:
            resp = api.create_media_list(project, [media_spec])
        except Exception as exc:
            raise RuntimeError("Could not create multiview") from exc

        thumbnail_info = None
        for progress, thumbnail_info in _upload_file(
            api, project, thumb_path, media_id=resp.id[0], filename=tiled_thumb_filename
        ):
            logger.info(f"Thumbnail upload progress: {progress}%")
        if not thumbnail_info:
            raise RuntimeError("Did not receive thumbnail info after upload")

        thumbnail_gif_info = None
        for progress, thumbnail_gif_info in _upload_file(
            api, project, thumb_gif_path, media_id=resp.id[0], filename=tiled_gif_filename
        ):
            logger.info(f"Thumbnail gif upload progress: {progress}%")
        if not thumbnail_gif_info:
            raise RuntimeError("Did not receive thumbnail gif info after upload")

        # Open images to get output resolution.
        thumb_image = Image.open(thumb_path)
        thumb_gif_image = Image.open(thumb_gif_path)

        # Create image definitions for thumbnails.
        defs = [
            {
                "path": thumbnail_info.key,
                "size": os.stat(thumb_path).st_size,
                "resolution": [thumb_image.height, thumb_image.width],
                "mime": f"image/{thumb_image.format.lower()}",
            },
            {
                "path": thumbnail_gif_info.key,
                "size": os.stat(thumb_gif_path).st_size,
                "resolution": [thumb_gif_image.height, thumb_gif_image.width],
                "mime": f"image/{thumb_gif_image.format.lower()}",
            },
        ]
        roles = ["thumbnail", "thumbnail_gif"]

        for img_def, role in zip(defs, roles):
            response = api.create_image_file(resp.id[0], role=role, image_definition=img_def)
            if not isinstance(response, MessageResponse):
                raise RuntimeError(f"Unexpected response type '{type(response)}'")

        # Add the multi definition.
        multi_def = {"layout": layout, "ids": media_ids}
        if quality:
            multi_def["quality"] = quality
        if frame_offset:
            if len(frame_offset) != n_ids:
                raise RuntimeError("Length of frame offsets did not match length of media IDs!")
            multi_def["frameOffset"] = frame_offset
        api.update_media(resp.id[0], {"multi": multi_def})

        return resp



def clone_section(src_api, src_section_id, dest_project, dest_api=None):
    """ Clone section.

    This can be used to clone sections from one project to another or from one
    host to another.

    Example for different host:

    .. code-block:: python

        src_api = tator.get_api(host, token)
        dest_api = tator.get_api(other_host, other_token)
        src_section_id = 1
        dest_project = 1
        response = tator.util.clone_section(src_api, src_section_id, dest_project, dest_api)
        print(response.message)

    Example for same host:

    .. code-block:: python

        api = tator.get_api(host, token)
        src_section_id = 1
        dest_project = 1
        response = tator.util.clone_section(src_api, src_section_id, dest_project)
        print(response.message)

    :param src_api: :class:`tator.TatorApi` object corresponding to source host or only
        host if this is a clone on same host.
    :param src_section_id: Unique integer identifying section to clone.
    :param dest_project: Unique integer identifying destination project.
    :param dest_api: :class:`tator.TatorApi` object corresponding to destination host.
    :returns: Response from section creation request.
    """
    section_obj = src_api.get_section(src_section_id)
    spec = {'name': section_obj.name}
    if section_obj.object_search:
        spec['object_search'] = section_obj.object_search
    if section_obj.tator_user_sections:
        spec['tator_user_sections'] = section_obj.tator_user_sections
    if dest_api is None:
        dest_api = src_api
    return dest_api.create_section(dest_project, section_spec=spec)



import math
import logging
from urllib.parse import urljoin

import requests

logger = logging.getLogger(__name__)

def _download_file(api, project, url, out_path):
    CHUNK_SIZE = 10 * 1024 * 1024
    MAX_RETRIES = 10
    # If this is a normal url, get headers.
    if url.startswith('/'):
        config = api.api_client.configuration
        host = config.host
        token = config.api_key['Authorization']
        prefix = config.api_key_prefix['Authorization']
        url = urljoin(host, url)
        # Supply token here for eventual media authorization
        headers = {
            'Authorization': f'{prefix} {token}',
            'Content-Type': f'application/json',
            'Accept-Encoding': 'gzip',
        }
    elif url.startswith('http'):
        headers = {}
    # If this is a S3 object key, get a download url.
    else:
        url = api.get_download_info(project, {'keys': [url]})[0].url
        headers = {}
    for attempt in range(MAX_RETRIES):
        try:
            with requests.get(url, stream=True, headers=headers) as r:
                r.raise_for_status()
                total_size = r.headers['Content-Length']
                total_chunks = math.ceil(int(total_size) / CHUNK_SIZE)
                chunk_count = 0
                last_progress = 0
                yield last_progress
                with open(out_path, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=CHUNK_SIZE):
                        if chunk:
                            chunk_count += 1
                            f.write(chunk)
                            this_progress = round((chunk_count / total_chunks) *100,1)
                            if this_progress != last_progress:
                                yield this_progress
                                last_progress = this_progress
                yield 100
            break
        except Exception as ex:
            logger.error(f"Failed to download {url} on attempt {attempt} {ex}...")
            pass




# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class User(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'elemental_id': 'str',
        'email': 'str',
        'first_name': 'str',
        'id': 'int',
        'is_staff': 'bool',
        'last_name': 'str',
        'profile': 'object',
        'username': 'str'
    }

    attribute_map = {
        'elemental_id': 'elemental_id',
        'email': 'email',
        'first_name': 'first_name',
        'id': 'id',
        'is_staff': 'is_staff',
        'last_name': 'last_name',
        'profile': 'profile',
        'username': 'username'
    }

    def __init__(self, elemental_id=None, email=None, first_name=None, id=None, is_staff=None, last_name=None, profile=None, username=None, local_vars_configuration=None):  # noqa: E501
        """User - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._elemental_id = None
        self._email = None
        self._first_name = None
        self._id = None
        self._is_staff = None
        self._last_name = None
        self._profile = None
        self._username = None
        self.discriminator = None

        if elemental_id is not None:
            self.elemental_id = elemental_id
        if email is not None:
            self.email = email
        if first_name is not None:
            self.first_name = first_name
        if id is not None:
            self.id = id
        if is_staff is not None:
            self.is_staff = is_staff
        if last_name is not None:
            self.last_name = last_name
        if profile is not None:
            self.profile = profile
        if username is not None:
            self.username = username

    @property
    def elemental_id(self):
        """
        Unique ID for a user across clusters/deployments

        :return: The elemental_id of this User. 
        :rtype: str
        """
        return self._elemental_id

    @elemental_id.setter
    def elemental_id(self, elemental_id):
        """
        Unique ID for a user across clusters/deployments

        :param elemental_id: The elemental_id of this User.
        :type: str
        """

        self._elemental_id = elemental_id

    @property
    def email(self):
        """
        Email address of user.

        :return: The email of this User. 
        :rtype: str
        """
        return self._email

    @email.setter
    def email(self, email):
        """
        Email address of user.

        :param email: The email of this User.
        :type: str
        """

        self._email = email

    @property
    def first_name(self):
        """
        First name of user.

        :return: The first_name of this User. 
        :rtype: str
        """
        return self._first_name

    @first_name.setter
    def first_name(self, first_name):
        """
        First name of user.

        :param first_name: The first_name of this User.
        :type: str
        """

        self._first_name = first_name

    @property
    def id(self):
        """
        Unique integer identifying a user.

        :return: The id of this User. 
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """
        Unique integer identifying a user.

        :param id: The id of this User.
        :type: int
        """

        self._id = id

    @property
    def is_staff(self):
        """
        Whether this user is a system administrator.

        :return: The is_staff of this User. 
        :rtype: bool
        """
        return self._is_staff

    @is_staff.setter
    def is_staff(self, is_staff):
        """
        Whether this user is a system administrator.

        :param is_staff: The is_staff of this User.
        :type: bool
        """

        self._is_staff = is_staff

    @property
    def last_name(self):
        """
        Last name of user.

        :return: The last_name of this User. 
        :rtype: str
        """
        return self._last_name

    @last_name.setter
    def last_name(self, last_name):
        """
        Last name of user.

        :param last_name: The last_name of this User.
        :type: str
        """

        self._last_name = last_name

    @property
    def profile(self):
        """
        User profile information

        :return: The profile of this User. 
        :rtype: object
        """
        return self._profile

    @profile.setter
    def profile(self, profile):
        """
        User profile information

        :param profile: The profile of this User.
        :type: object
        """

        self._profile = profile

    @property
    def username(self):
        """
        Username of user.

        :return: The username of this User. 
        :rtype: str
        """
        return self._username

    @username.setter
    def username(self, username):
        """
        Username of user.

        :param username: The username of this User.
        :type: str
        """

        self._username = username

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, User):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, User):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class MediaStats(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'count': 'int',
        'download_size': 'int',
        'duration': 'float',
        'total_size': 'int'
    }

    attribute_map = {
        'count': 'count',
        'download_size': 'download_size',
        'duration': 'duration',
        'total_size': 'total_size'
    }

    def __init__(self, count=None, download_size=None, duration=None, total_size=None, local_vars_configuration=None):  # noqa: E501
        """MediaStats - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._count = None
        self._download_size = None
        self._duration = None
        self._total_size = None
        self.discriminator = None

        if count is not None:
            self.count = count
        if download_size is not None:
            self.download_size = download_size
        if duration is not None:
            self.duration = duration
        if total_size is not None:
            self.total_size = total_size

    @property
    def count(self):
        """

        :return: The count of this MediaStats. 
        :rtype: int
        """
        return self._count

    @count.setter
    def count(self, count):
        """

        :param count: The count of this MediaStats.
        :type: int
        """
        if (self.local_vars_configuration.client_side_validation and
                count is not None and count < 0):  # noqa: E501
            raise ValueError("Invalid value for `count`, must be a value greater than or equal to `0`")  # noqa: E501

        self._count = count

    @property
    def download_size(self):
        """

        :return: The download_size of this MediaStats. 
        :rtype: int
        """
        return self._download_size

    @download_size.setter
    def download_size(self, download_size):
        """

        :param download_size: The download_size of this MediaStats.
        :type: int
        """
        if (self.local_vars_configuration.client_side_validation and
                download_size is not None and download_size < 0):  # noqa: E501
            raise ValueError("Invalid value for `download_size`, must be a value greater than or equal to `0`")  # noqa: E501

        self._download_size = download_size

    @property
    def duration(self):
        """

        :return: The duration of this MediaStats. 
        :rtype: float
        """
        return self._duration

    @duration.setter
    def duration(self, duration):
        """

        :param duration: The duration of this MediaStats.
        :type: float
        """
        if (self.local_vars_configuration.client_side_validation and
                duration is not None and duration < 0):  # noqa: E501
            raise ValueError("Invalid value for `duration`, must be a value greater than or equal to `0`")  # noqa: E501

        self._duration = duration

    @property
    def total_size(self):
        """

        :return: The total_size of this MediaStats. 
        :rtype: int
        """
        return self._total_size

    @total_size.setter
    def total_size(self, total_size):
        """

        :param total_size: The total_size of this MediaStats.
        :type: int
        """
        if (self.local_vars_configuration.client_side_validation and
                total_size is not None and total_size < 0):  # noqa: E501
            raise ValueError("Invalid value for `total_size`, must be a value greater than or equal to `0`")  # noqa: E501

        self._total_size = total_size

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, MediaStats):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, MediaStats):
            return True

        return self.to_dict() != other.to_dict()



#!/usr/bin/env python

import argparse
import logging
import subprocess
import json
import os
import sys
import psutil
import time

from ..util.get_api import get_api
from ..util._upload_file import _upload_file
from ..openapi.tator_openapi.models import MessageResponse

from .make_fragment_info import make_fragment_info

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# If HW is available, use this as lookup swap
encoder_lookup=None

SW_TO_HW_PIXEL_FORMAT_CONVERSION = {
    'yuv420p': 'nv12', # QSV h264 supports this
    'yuv420p10le': 'p010le', # QSV h264 supports this
    'yuv422p': 'nv16', # Not sure if any hw encoders actually support this
    'yuv444p' : 'nv21' # Not sure if any hw encoders actually support this
}

def _get_resource_usage():
    """ Returns an object representing memory + CPU usage """
    cpu_percentages=psutil.cpu_percent(percpu=True)
    load_avg = psutil.getloadavg()
    memory_raw = psutil.virtual_memory()
    memory_utilization = (memory_raw.used / memory_raw.total) * 100
    memory_in_gb = memory_raw.used / 1024 / 1024
    return {'cpu_percentages': cpu_percentages,
           'load_avg': load_avg,
           'memory_utilization': memory_utilization,
           'memory_in_gb': memory_in_gb}
def _launch_and_monitor_resources(cmd, interval=5):
    proc = subprocess.Popen(cmd)
    start = time.time()
    while proc.returncode == None:
        try:
            proc.wait(timeout=interval)
        except subprocess.TimeoutExpired:
            pass # we don't care
        logger.info(f"RESOURCE_INFO = {json.dumps(_get_resource_usage())}")
    if proc.returncode != 0:
        raise RuntimeError("Transcode process failed")
    end = time.time()
    logger.info(f"cmd={cmd}")
    logger.info(f"time={end-start}")

def find_best_encoder(codec, hwaccel=False):
    """ Find the best encoder based on what is available on the system """
    global encoder_lookup
    if encoder_lookup is None:
        # Default codecs
        encoder_lookup={"hevc": os.getenv("TATOR_HEVC_ENCODER", "libx265"),
                        "h264": os.getenv("TATOR_H264_ENCODER", "libx264"),
                        "av1": os.getenv("TATOR_AV1_ENCODER", "libsvtav1")}
        if hwaccel:
            cmd = [
                "ffmpeg",
                "-encoders" ]
            output=subprocess.run(cmd,stdout=subprocess.PIPE,check=True).stdout.decode()
            # Look for QSV, but use VAAPI frontend
            # TODO: Use `vainfo` directly to query available hardware entry points
            if output.find("libsvt_hevc") >= 0:
                encoder_lookup["hevc"] = "libsvt_hevc"
            if output.find("hevc_qsv") >= 0:
                encoder_lookup["hevc"] = "hevc_vaapi"
            if output.find("h264_qsv") >= 0:
                encoder_lookup["h264"] = "h264_vaapi"
            if output.find("av1_qsv") >= 0:
                encoder_lookup["av1"] = "av1_vaapi"
        logger.info("encoder_lookup = %s", encoder_lookup)
    return encoder_lookup.get(codec, codec)

def parse_args():
    parser = argparse.ArgumentParser(description='Transcodes a raw video.')
    parser.add_argument('--url', type=str, help='URL where original file is hosted.')
    parser.add_argument('--work_dir', type=str, help='Directory where info should be saved.')
    parser.add_argument('--host', type=str, default='https://cloud.tator.io', help='Host URL.')
    parser.add_argument('--token', type=str, help='REST API token.')
    parser.add_argument('--media', type=int, help='Unique integer identifying a media.')
    parser.add_argument('--category', required=True, help='One of streaming, archival, or audio.')
    parser.add_argument('--raw_width', type=int, help='Pixel width of original video.')
    parser.add_argument('--raw_height', type=int, help='Pixel height of original video.')
    parser.add_argument('--configs', type=str, help='Comma separated list of output configs, '
                                                    'format is resolution:crf:codec.')
    parser.add_argument('--size', type=int, help='Size of the file, if not inferrable')
    parser.add_argument('--hwaccel', action='store_true', help="Use hardware acceleration.")
    parser.add_argument('--force_fps', type=float, default=-1, help='Force a specific fps for the video.')
    return parser.parse_args()

def get_length_of_file(path):
    cmd = [
        "ffprobe",
        "-v","error",
        "-show_entries", "stream:format=duration",
        "-print_format", "json",
        "-select_streams", "v",
        path,
    ]
    output = subprocess.run(cmd, stdout=subprocess.PIPE, check=True).stdout
    video_info = json.loads(output)
    stream_idx=0
    for idx, stream in enumerate(video_info["streams"]):
        if stream["codec_type"] == "video":
            stream_idx=idx
            break
    stream = video_info["streams"][stream_idx]
    stream = {**video_info.get("format", {}), **stream}
    fps, length = get_length_info(stream)
    return fps, length

def make_video_definition(path, size=None):
    cmd = [
        "ffprobe",
        "-v","error",
        "-show_entries", "stream",
        "-print_format", "json",
        "-select_streams", "v",
        path,
    ]
    output = subprocess.run(cmd, stdout=subprocess.PIPE, check=True).stdout
    video_info = json.loads(output)
    stream_idx=0
    if size is None or size <= 0:
        size = os.stat(path).st_size
    for idx, stream in enumerate(video_info["streams"]):
        if stream["codec_type"] == "video":
            stream_idx=idx
            break
    stream = video_info["streams"][stream_idx]
    video_def = {"resolution": (stream["height"], stream["width"]),
                 "codec": stream["codec_name"],
                 "codec_description": stream["codec_long_name"],
                 "size": size,
                 "bit_rate": int(stream.get("bit_rate",-1))}
    return video_def

def convert_streaming(host, token, media, path, outpath, raw_width, raw_height, configs, hwaccel=False, inhibit_upload=False, force_fps=-1):
    logger.info("Transcoding %s to %s...", path, outpath)
    # Get workload parameters.
    os.makedirs(outpath, exist_ok=True)

    # Convert settings into resolution/crf/codec/presets/pixel_formats.
    resolutions = [int(config.split(':')[0]) for config in configs]
    crfs = [config.split(':')[1] for config in configs]
    codecs = [config.split(':')[2] for config in configs]
    presets = [config.split(':')[3] for config in configs]

    # Handle pixel format addition as an optional argument
    # This will maintain backwards API compatibility to folks using
    # this function
    # TODO: It'd be nice if we didn't pass in the config so awkwardly.
    pixel_formats = []
    for c in configs:
        split_comps = c.split(':')
        if len(split_comps) == 5:
            pixel_formats.append(split_comps[4])
        else:
            pixel_formats.append("yuv420p")

    # Need to get avg_framerate
    if isinstance(path, list):
        # Assume that files are the same frame rate and resolution
        ffprobe_path = path[0]
    else:
        ffprobe_path = path

    cmd = [
        "ffprobe",
        "-v","error",
        "-show_entries", "stream",
        "-print_format", "json",
        "-select_streams", "v",
        ffprobe_path,
    ]
    output = subprocess.run(cmd, stdout=subprocess.PIPE, check=True).stdout
    video_info = json.loads(output)
    if force_fps < 0:
        avg_frame_rate=video_info['streams'][0]['avg_frame_rate']
        fps_fractional = avg_frame_rate.split("/")
        avg_frame_rate = float(fps_fractional[0]) / float(fps_fractional[1])
        avg_frame_rate = round(avg_frame_rate, 4)
    else:
        avg_frame_rate=force_fps
    tags=video_info['streams'][0].get('tags')
    rotation = 0
    if tags is not None:
        rotation = int(tags.get('rotate', "0")) % 360
    transpose = "null"
    if rotation == 90:
        transpose = "transpose=2"
    elif rotation == 180:
        transpose = "hflip,vflip"
    elif rotation == 270:
        transpose = "transpose=1"

    vid_dims = [raw_height, raw_width]
    cmd = [
        "ffmpeg", "-y",
        "-loglevel", "error",
        "-progress", "-",
        "-stats_period", "10"
    ]
    # We are going to concatenate multiple segments together
    if type(path) == list:
        num_segments = len(path)
        for segment in path:
            cmd.extend(["-noautorotate", "-i", segment])
    else:
        num_segments = 1
        cmd.extend(["-noautorotate", "-i", path])

    cmd.extend(["-noautorotate", "-i", os.path.join(os.path.dirname(os.path.abspath(__file__)), "black.mp4")])
    

    vaapi_present = [c for c in codecs if find_best_encoder(c, hwaccel).find('vaapi') >= 0]
    if vaapi_present:
        cmd.extend(['-init_hw_device', 'vaapi=hw',
                    '-filter_hw_device', 'hw'])

    logger.info("Transcoding to %s", resolutions)
    for ridx, resolution in enumerate(resolutions):
        per_res = ["-an",
            "-metadata:s", "handler_name=tator",
            "-g", "25",
            "-movflags",
            "faststart+frag_keyframe+empty_moov+default_base_moof"]
        logger.info(f"Generating resolution @ {resolution}")
        output_file = os.path.join(outpath, f"{resolution}.mp4")
        codec = find_best_encoder(codecs[ridx], hwaccel)
        quality_flag = "-crf"
        pixel_format = pixel_formats[ridx]
        hw_upload = ''
        preset = presets[ridx]
        if codec.find("qsv") >= 0:
            quality_flag = "-global_quality"
        if codec.find("vaapi") >= 0:
            quality_flag = "-global_quality"
            pixel_format = SW_TO_HW_PIXEL_FORMAT_CONVERSION[pixel_format]
            # add format filter for vaapi + add hwupload to incantation
            hw_upload=f'[uf{ridx}];[uf{ridx}]format={pixel_format}[format{ridx}];[format{ridx}]hwupload'

        if codec.find('264') > 0:
            preset = preset if preset else 'fast'
            per_res.extend(["-preset", preset,
                            "-tune", "fastdecode"])
        if codec.find('libsvtav1') >= 0:
            preset = preset if preset else '5'
            per_res.extend(['-preset', preset, "-tune", "fastdecode"])

        cmd.extend([*per_res,
                    "-vcodec", codec,
                    "-pix_fmt", pixel_format,
                    quality_flag, crfs[ridx],
                    "-filter_complex"
                    ])
        
        # Construct the complex filter
        if num_segments == 1:
            cmd.extend([
                    # Scale the black mp4 to the input resolution prior to concating and scaling back down.
                    f"[0:v:0]{transpose}[rot0];[rot0]yadif[a{ridx}];[a{ridx}]setsar=1[vid{ridx}];[1:v:0]scale={vid_dims[1]}:{vid_dims[0]},setsar=1[bv{ridx}];[vid{ridx}][bv{ridx}]concat=n=2:v=1:a=0[rv{ridx}];[rv{ridx}]scale=-2:{resolution}[catv{ridx}];[catv{ridx}]pad=ceil(iw/2)*2:ceil(ih/2)*2[norate{ridx}];[norate{ridx}]fps={avg_frame_rate}{hw_upload}[outv{ridx}]",
                    "-metadata:s:v:0", "rotate=0",
                    "-map", f"[outv{ridx}]",
                    output_file])
        else:
            filter_string = ""
            for seg_idx in range(num_segments):
                filter_string += f"[{seg_idx}:v:0]{transpose}[rot{seg_idx}];[rot{seg_idx}]yadif[a{seg_idx}_{ridx}];[a{seg_idx}_{ridx}]setsar=1[vid{seg_idx}_{ridx}];"
                
            filter_string += f"[{num_segments}:v:0]scale={vid_dims[1]}:{vid_dims[0]},setsar=1[bv{ridx}];"
            filter_string += "".join([f"[vid{seg_idx}_{ridx}]" for seg_idx in range(num_segments)]) 
                
            filter_string += f"[bv{ridx}]concat=n={num_segments+1}:v=1:a=0[rv{ridx}];[rv{ridx}]scale=-2:{resolution}[catv{ridx}];[catv{ridx}]pad=ceil(iw/2)*2:ceil(ih/2)*2[norate{ridx}];[norate{ridx}]fps={avg_frame_rate}{hw_upload}[outv{ridx}]"

            cmd.extend([filter_string,
                        "-metadata:s:v:0", "rotate=0",
                        "-map", f"[outv{ridx}]",
                        output_file])

    logger.info('ffmpeg cmd = {}'.format(cmd))
    _launch_and_monitor_resources(cmd)

    api = get_api(host, token)
    media_obj = api.get_media(media)

    # Skip this check for now, need to calculate concat duration for concat files
    '''
    for ridx, resolution in enumerate(resolutions):
        output_file = os.path.join(outpath, f"{resolution}.mp4")
        _, res_length = get_length_of_file(output_file)
        length_delta = abs((media_obj.num_frames - res_length)/media_obj.num_frames)
        assert length_delta < 0.1 # Assert length delta is less than 10 percent.
    '''
    
    # if we are inhibiting upload, we are done!
    if inhibit_upload:
        return

    for resolution in resolutions:
        output_file = os.path.join(outpath, f"{resolution}.mp4")

        segments_file = os.path.join(outpath, f"{resolution}.json")
        make_fragment_info(output_file, segments_file)

        logger.info("Uploading transcoded file...")
        for progress, upload_info in _upload_file(api, media_obj.project, output_file,
                                                  media_id=media, filename=os.path.basename(output_file)):
            logger.info(f"Progress: {progress}%")

        logger.info("Uploading segments file...")
        for progress, segment_info in _upload_file(api, media_obj.project, segments_file,
                                                  media_id=media, filename=os.path.basename(segments_file)):
            logger.info(f"Progress: {progress}%")

        # Construct video definition.
        video_def = {
            **make_video_definition(output_file),
            'path': upload_info.key,
            'segment_info': segment_info.key,
        }

        # Patch in video file with the api.
        response = api.create_video_file(media, role='streaming', video_definition=video_def)
        assert isinstance(response, MessageResponse)

def default_archival_upload(api, host, media, path, encoded, size):
    # Default action if no archive config is upload raw video.
    media_obj = api.get_media(media)
    logger.info(f"Uploading original file as archival...")
    for progress, upload_info in _upload_file(api, media_obj.project, path,
                                              media_id=media, filename=os.path.basename(path), file_size=size):
        logger.info(f"Progress: {progress}%")
    video_def = make_video_definition(path, size)

    # If video was encoded, set codec_mime to video/mp4; otherwise do
    # not set codec_mime.
    if encoded:
        video_def['codec_mime'] = 'video/mp4'

    # Patch in video file with the api.
    video_def['path'] = upload_info.key
    response = api.create_video_file(media, role='archival', video_definition=video_def)
    assert isinstance(response, MessageResponse)


def convert_archival(
    host,
    token,
    media,
    path,
    outpath,
    raw_width,
    raw_height,
    size=None,
    explicit_config=None,
    hwaccel=False,
    inhibit_upload=False,
):
    # Retrieve this media's type to inspect archive config.
    api = get_api(host, token)
    media_obj = api.get_media(media)
    media_type = api.get_media_type(media_obj.type)
    config = explicit_config or media_type.archive_config
    if config is not None:
        for idx, archive_config in enumerate(config):
            os.makedirs(outpath, exist_ok=True)
            output_file = os.path.join(outpath, f"archival_{idx}.mp4")
            if archive_config.encode.vcodec == 'copy':
                # If no encode, just use the original file.
                output_file = path
            else:
                # Encode the media to archival format.
                codec = find_best_encoder(archive_config.encode.vcodec, hwaccel)
                quality_flag = "-crf"
                pixel_format = archive_config.encode.pixel_format
                tune_settings = ["-preset", archive_config.encode.preset,
                                 "-tune", archive_config.encode.tune]
                if archive_config.encode.movflags:
                    tune_settings.extend(['-movflags', archive_config.encode.movflags])
                if codec.find("qsv") >= 0:
                    quality_flag = "-global_quality"
                    pixel_format = SW_TO_HW_PIXEL_FORMAT_CONVERSION[pixel_format]
                    tune_settings=[] #QSV doesn't do tuning
                elif codec == "libsvt_hevc":
                    # SVT for HEVC does not do tuning or CRF
                    tune_settings=[]
                    quality_flag = "-global_quality"

                hw_preamble = []
                vaapi_present = codec.find('vaapi') >= 0
                if vaapi_present:
                    hw_preamble = ['-init_hw_device', 'vaapi=hw',
                                   '-filter_hw_device', 'hw']
                cmd = [
                    "ffmpeg",
                    "-loglevel", "error",
                    "-progress", "-",
                    "-stats_period", "10",
                    *hw_preamble,
                    "-y",
                    "-i", path,
                    "-vcodec", codec,
                    "-vf", "yadif",
                    quality_flag, str(archive_config.encode.crf),
                    "-pix_fmt", pixel_format,
                    *tune_settings
                ]

                if archive_config.encode.vcodec == 'hevc':
                    cmd += ["-tag:v", "hvc1"]
                elif archive_config.encode.vcodec == 'h264':
                    cmd += ["-tag:v", "avc1"]
                cmd.append(output_file)

                logger.info('ffmpeg cmd = {}'.format(cmd))
                _launch_and_monitor_resources(cmd)

            if inhibit_upload:
                continue
            if archive_config.s3_storage is None:
                default_archival_upload(api, host, media, output_file, True, size)
            else:
                import boto3
                # Get credentials from config object.
                aws_access_key = archive_config.s3_storage.aws_access_key
                aws_secret_access_key = archive_config.s3_storage.aws_secret_access_key
                bucket_name = archive_config.s3_storage.bucket_name
                logger.info(f"Uploading {output_file} to S3 bucket {bucket_name}...")

                # Upload the video to S3.
                client = boto3.client('s3', aws_access_key_id=aws_access_key,
                                      aws_secret_access_key=aws_secret_access_key)
                client.upload_file(output_file, bucket_name, os.path.basename(output_file))


def make_audio_definition(disk_file):
    cmd = [
        "ffprobe",
        "-v","error",
        "-show_entries", "stream",
        "-print_format", "json",
        "-select_streams", "a",
        disk_file,
    ]
    output = subprocess.run(cmd, stdout=subprocess.PIPE, check=True).stdout
    audio_info = json.loads(output)
    stream_idx=0
    for idx, stream in enumerate(audio_info["streams"]):
        if stream["codec_type"] == "audio":
            stream_idx=idx
            break
    stream = audio_info["streams"][stream_idx]
    audio_def = {"codec": stream["codec_name"],
                 "codec_description": stream["codec_long_name"],
                 "size": os.stat(disk_file).st_size,
                 "bit_rate": int(stream.get("bit_rate",-1))}
    return audio_def


def convert_audio(host, token, media, path, outpath, inhibit_upload=False):
    if isinstance(path, list):
        logger.info("Concatenation with audio not supported yet")
        return
    os.makedirs(outpath, exist_ok=True)
    logger.info("Extracting audio")
    output_file = os.path.join(outpath, f"audio.m4a")
    audio_extraction=["ffmpeg", "-y",
                      "-loglevel", "error",
                      "-progress", "-",
                      "-stats_period", "10",
                      "-i", path,
                      "-vn", # Strip video
                      "-c:a", "aac",
                      "-ac", "2",
                      output_file]
    subprocess.run(audio_extraction, check=True)
    logger.info("Finished extracting audio!")

    if inhibit_upload:
        return

    # Upload audio.
    api = get_api(host, token)
    media_obj = api.get_media(media)
    for progress, upload_info in _upload_file(api, media_obj.project, output_file,
                                              media_id=media, filename=os.path.basename(output_file)):
        logger.info(f"Progress: {progress}%")

    # Patch in audio file with the api.
    audio_def = {**make_audio_definition(output_file),
                 'path': upload_info.key}
    response = api.create_audio_file(media, role='audio', audio_definition=audio_def)
    assert isinstance(response, MessageResponse)


def get_length_info(stream):
    """ Given a json dump of the stream return the length of the video """
    fps_fractional = stream["avg_frame_rate"].split("/")
    fps = float(fps_fractional[0]) / float(fps_fractional[1])

    if 'duration' in stream:
        seconds = float(stream["duration"])
    elif 'tags' in stream and 'DURATION' in stream['tags']:
        length = stream['tags']['DURATION'].split(':')
        seconds = float(length[0])*3600
        seconds += float(length[1])*60
        seconds += float(length[2])
    else:
        raise Exception('No way to determine file length!')

    num_frames = float(fps * seconds)
    return fps,int(num_frames)

if __name__ == '__main__':
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)

    # Parse arguments.
    args = parse_args()

    if args.category == 'streaming':
        if args.configs == '':
            configs = []
        else:
            configs = [res for res in args.configs.split(',')]
        convert_streaming(args.host, args.token, args.media, args.url, args.work_dir,
                          args.raw_width, args.raw_height, configs, args.hwaccel, False, args.force_fps)
    elif args.category == 'archival':
        convert_archival(args.host, args.token, args.media, args.url, args.work_dir,
                         args.raw_width, args.raw_height, args.size)
    elif args.category == 'audio':
        convert_audio(args.host, args.token, args.media, args.url, args.work_dir)



import os
import math
import tempfile
import logging
from uuid import uuid1
from collections import defaultdict
from urllib.parse import urlparse, urljoin

import progressbar

from .md5sum import md5sum
from ._download_file import _download_file
from ._upload_file import _upload_file

logger = logging.getLogger(__name__)

class HostTransfer:
    def __init__(self, src_api, src_project, dest_api, dest_project):
        """ Sets up authentication for transferring file.

        :param src_api: :class:`tator.TatorApi` object corresponding to source host.
        :param src_project: Unique integer identifying the source project.
        :param dest_api: :class:`tator.TatorApi` object corresponding to destination host.
        :param dest_project: Unique integer identifying the destination project.
        """
        # Set up source.
        self.src_api = src_api
        self.src_project = src_project

        # Set up destination.
        self.dest_api = dest_api
        self.dest_project = dest_project

    def __call__(self, src_url, media_id=None, return_url=False):
        """ Downloads a file from one host and uploads it to another.

        :param src_url: Relative URL of source file to download.
        :param media_id: Destination media ID.
        :param return_url: Return download URL if true.
        :returns: URL or path of destination file.
        """
        parsed = urlparse(src_url)
        filename = os.path.basename(parsed.path)
        with tempfile.TemporaryDirectory() as temp_dir:
            temp = os.path.join(temp_dir, filename)
            print(f"Downloading {filename}")
            bar = progressbar.ProgressBar(max_value=100)
            for progress in _download_file(self.src_api, self.src_project, src_url, temp):
                bar.update(progress)
            bar.update(100)
            print(f"Uploading {filename}")
            bar = progressbar.ProgressBar(max_value=100)
            for progress, upload_info in _upload_file(self.dest_api, self.dest_project, temp,
                                               media_id=media_id, filename=filename):
                bar.update(progress)
            bar.update(100)
            out = upload_info.key
        if return_url:
            out = self.dest_api.get_download_info(self.dest_project,
                                                  {'keys': [upload_info.key]})[0].url
        return out

def clone_media_list(src_api, query_params, dest_project, media_mapping={}, dest_type=-1,
                     dest_section='', dest_api=None, ignore_transfer=False):
    """ Clone media list.

    This can be used to clone media from one project to another or from one
    host to another. In the case of the same host, the media files are not
    copied. If the destination host is different (`dest_api` is a :class:`tator.TatorApi`
    object), the transcoded files for each media object is downloaded and then
    uploaded to the other host, and a new :class:`tator.Media` object is created.

    Example for different host:

    .. code-block:: python

        src_api = tator.get_api(host, token)
        dest_api = tator.get_api(other_host, other_token)
        query_params = {'project': 1, 'media_id': [1]}
        media_mapping = {} # Only required if query contains multi media types
        dest_project = 1
        dest_type = -1
        dest_section = 'My cloned media'
        created_ids = []
        generator = clone_media_list(src_api, query_params, dest_project, media_mapping,
                                     dest_type, dest_section, dest_api)
        for num_created, num_total, response, id_map in generator:
            print(f"Created {num_created} of {num_total} files...")
            created_ids.append(response.id[0])
        print(f"Finished creating {num_created} files!")

    Example for same host:

    .. code-block:: python

        api = tator.get_api(host, token)
        query_params = {'media_id': [1]}
        dest_project = 1
        created_ids = []
        generator = clone_media_list(src_api, query_params, dest_project)
        for num_created, num_total, response, id_map in generator:
            print(f"Created {num_created} of {num_total} files...")
            created_ids += response.id[0] # This response is from the CloneMedia endpoint.
        print(f"Finished creating {num_created} files!")

    :param src_api: :class:`tator.TatorApi` object corresponding to source host or only
        host if this is a clone on same host.
    :param query_params: Dictionary containing query parameters for source media list.
    :param dest_project: Unique integer identifying destination project.
    :param media_mapping: Mapping between source and destination media IDs. Only used
        if the media list contains multi media types and the clone is being done for
        different hosts.
    :param dest_type: Unique integer identifying destination media type. If set to
        -1, the media type is set to the first media type in the project.
    :param dest_section: Name of destination section.
    :param dest_api: :class:`tator.TatorApi` object corresponding to destination host.
    :param ignore_transfer: True if media files should not be transferred. Media object will still be created.
        Paths in media_files will be invalid.
    :returns: Generator containing number of files created, number of files total,
        most recent response from media creation operation, and mapping between original IDs
        and created IDs.
    """
    # Make sure query has a project.
    if 'project' not in query_params:
        raise Exception("Query parameters must include a project!")
    src_project = query_params['project']

    # Guess the media type if it was not given.
    if dest_type == -1:
        if dest_api is None:
            media_types = src_api.get_media_type_list(dest_project)
        else:
            media_types = dest_api.get_media_type_list(dest_project)
        if len(media_types) == 0:
            raise Exception('Specified project does not have any media types!')
        dest_type = media_types[0].id

    # Get medias.
    try:
        medias = src_api.get_media_list(**query_params, presigned=86400)
    except:
        # Support legacy api
        medias = src_api.get_media_list(**query_params)
    total_files = len(medias)

    # Clone the media.
    created_ids = []
    if (dest_api is None) or (dest_api is src_api):
        for idx in range(0, len(medias), 100): # Go in batches of 100
            # Clone media locally.
            media_ids = [media.id for media in medias[idx:idx+100]]
            response = src_api.clone_media_list(src_project,
                                                media_id=media_ids,
                                                clone_media_spec={
                                                    'dest_project': dest_project,
                                                    'dest_section': dest_section,
                                                    'dest_type': dest_type,
                                                })
            created_ids += response.id
            id_map = {src_id: dest_id for src_id, dest_id in zip(media_ids, response.id)}
            yield (len(created_ids), total_files, response, id_map)
    else:
        # Clone media to another host.
        if not ignore_transfer:
            transfer = HostTransfer(src_api, src_project, dest_api, dest_project)
        for media in medias:
            # Set up media spec.
            attributes = media.attributes
            if 'tator_user_sections' in attributes:
                attributes.pop('tator_user_sections', None)
            media_spec = {
                'type': dest_type,
                'name': media.name,
                'md5': media.md5,
                'fps': media.fps,
                'num_frames': media.num_frames,
                'codec': media.codec,
                'width': media.width,
                'height': media.height,
                'attributes': attributes,
                'section': dest_section if dest_section else media.section,
            }
            if media.gid:
                media_spec['gid'] = media.gid
            if media.uid:
                media_spec['uid'] = media.uid

            # Create the media object.
            response = dest_api.create_media_list(dest_project, body=[media_spec])
            id_map = {media.id: response.id[0]}

            # Transfer videos.
            if media.media_files:
                media_files = media.media_files.to_dict()
                for key in ['streaming', 'archival', 'audio', 'image', 'thumbnail', 'thumbnail_gif']:
                    if ignore_transfer:
                        continue
                    if media_files.get(key):
                        for item in media_files[key]:
                            media_def = {k: v for k, v in item.items()
                                         if v is not None}
                            src_path = media_def.pop('path', None)
                            media_def['path'] = transfer(src_path, media_id=response.id[0])
                            if key == 'streaming':
                                src_segments = media_def.pop('segment_info', None)
                                media_def['segment_info'] = transfer(src_segments, media_id=response.id[0])
                            if key in ['streaming', 'archival']:
                                dest_api.create_video_file(response.id[0], role=key,
                                                           video_definition=media_def)
                            elif key in ['image', 'thumbnail', 'thumbnail_gif']:
                                dest_api.create_image_file(response.id[0], role=key,
                                                           image_definition=media_def)
                            elif key in ['audio']:
                                dest_api.create_audio_file(response.id[0], role=key,
                                                           audio_definition=media_def)
                if media.media_files.ids:
                    dest_ids = []
                    for id_ in media.media_files.ids:
                        if id_ in media_mapping:
                            dest_ids.append(media_mapping[id_])
                        else:
                            raise Exception(f"Source media ID {id_} does not exist in media "
                                             "mapping! Individual videos should be migrated "
                                             "before multi videos.")
                    update = {'multi': {'ids': dest_ids}}
                    if media.media_files.layout:
                        update['multi']['layout'] = media.media_files.layout
                    if media.media_files.quality:
                        update['multi']['quality'] = media.media_files.quality
                    dest_api.update_media(response.id[0], media_update=update)
            created_ids.append(response.id[0])
            yield (len(created_ids), total_files, response, id_map)



from pprint import pformat
import datetime
import random
from time import sleep
import uuid
from collections import Counter
import pytest

import tator
from ._common import assert_close_enough


def wait_for_parity(tator_api, project, patch, expected_ids):
    attribute_filter = [f"test_string::{patch['attributes']['test_string']}"]
    state_id_query = {"ids": expected_ids}
    total_timeout = 60.0  # seconds
    wait_time = 2.0  # seconds

    print(f"Using attribute filter:\n{pformat(attribute_filter)}")
    from_psql = tator_api.get_state_list_by_id(
        project,
        state_id_query=state_id_query,
        attribute=attribute_filter,
    )
    count = tator_api.get_state_count_by_id(
        project,
        state_id_query=state_id_query,
        attribute=attribute_filter,
    )
    assert len(from_psql) == count
    for idx in range(int(total_timeout / wait_time) + 1):
        from_es = tator_api.get_state_list_by_id(
            project, state_id_query=state_id_query, attribute=attribute_filter
        )
        count = tator_api.get_state_count_by_id(
            project, state_id_query=state_id_query, attribute=attribute_filter
        )
        assert len(from_es) == count

        if len(from_es) == len(from_psql):
            print(f"Found expected number of results after {idx * wait_time} seconds")
            return True

        sleep(wait_time)

    print(f"Did not find expected number of results after {idx * wait_time} seconds")
    return False


def random_state(project, state_type, video_obj, post=False):
    attributes = {
        "test_bool": random.choice([False, True]),
        "test_int": random.randint(-1000, 1000),
        "test_float": random.uniform(-1000.0, 1000.0),
        "test_enum": random.choice(["a", "b", "c"]),
        "test_string": str(uuid.uuid1()),
        "test_datetime": datetime.datetime.now().isoformat(),
        "test_geopos": [random.uniform(-180.0, 180.0), random.uniform(-90.0, 90.0)],
        "test_float_array": [random.uniform(-1.0, 1.0) for _ in range(3)],
    }
    out = {
        "project": project,
        "type": state_type,
        "media_ids": [video_obj.id],
        "frame": random.randint(0, video_obj.num_frames - 1),
        "attributes": attributes,
    }
    return {**out}


def comparison_query(tator_api, project, state_ids, exclude):
    """Runs a random query and compares results with ES enabled and disabled."""
    bool_value = random.choice([True, False])
    int_lower = random.randint(-1000, 0)
    int_upper = random.randint(0, 1000)
    float_lower = random.uniform(-1000.0, 0.0)
    float_upper = random.uniform(0.0, 1000.0)
    enum_value = random.choice(["a", "b", "c"])
    state_id_query = {"ids": state_ids}
    attribute_filter = [
        f"test_bool::{'true' if bool_value else 'false'}",
        f"test_enum::{enum_value}",
    ]
    attribute_lte_filter = [f"test_int::{int_upper}"]
    attribute_gte_filter = [f"test_int::{int_lower}"]
    attribute_lt_filter = [f"test_float::{float_upper}"]
    attribute_gt_filter = [f"test_float::{float_lower}"]
    print("Starting PSQL query...")
    t0 = datetime.datetime.now()
    from_psql = tator_api.get_state_list_by_id(
        project,
        state_id_query=state_id_query,
        attribute=attribute_filter,
        attribute_lte=attribute_lte_filter,
        attribute_gte=attribute_gte_filter,
        attribute_lt=attribute_lt_filter,
        attribute_gt=attribute_gt_filter,
    )
    psql_time = datetime.datetime.now() - t0
    print("Starting ES query...")
    t0 = datetime.datetime.now()
    from_es = tator_api.get_state_list_by_id(
        project,
        state_id_query=state_id_query,
        attribute=attribute_filter,
        attribute_lte=attribute_lte_filter,
        attribute_gte=attribute_gte_filter,
        attribute_lt=attribute_lt_filter,
        attribute_gt=attribute_gt_filter,
        force_es=1,
    )
    es_time = datetime.datetime.now() - t0

    print("Checking PSQL and ES ids...")
    psql_ids = [ele.id for ele in from_psql]
    es_ids = [ele.id for ele in from_es]
    assert Counter(psql_ids) == Counter(es_ids)

    print("Checking PSQL and ES values...")
    assert len(from_psql) == len(from_es)
    for psql, es in zip(from_psql, from_es):
        assert_close_enough(psql, es, exclude)
        assert psql.attributes["test_bool"] == bool_value
        assert psql.attributes["test_int"] <= int_upper
        assert psql.attributes["test_int"] >= int_lower
        assert psql.attributes["test_float"] < float_upper
        assert psql.attributes["test_float"] > float_lower
        assert psql.attributes["test_enum"] == enum_value
    return psql_time, es_time


def test_state_crud(host, token, project, video_type, empty_video, state_type):
    video = empty_video
    tator_api = tator.get_api(host, token)
    video_obj = tator_api.get_media(video)

    # These fields will not be checked for object equivalence after patch.
    exclude = ["project", "type", "media_ids", "id", "type", "user", "frame", "ids", "in_place"]

    # Test bulk create.
    num_states = random.randint(2000, 10000)
    states = [random_state(project, state_type, video_obj, post=True) for _ in range(num_states)]
    state_ids = []
    for response in tator.util.chunked_create(tator_api.create_state_list, project, body=states):
        state_ids += response.id
    assert len(state_ids) == len(states)
    print(f"Created {len(state_ids)} states!")

    # Verify list contains the number of entities created
    response = tator_api.get_state_list(project, type=state_type)
    assert len(response) == num_states

    # Test media retrieval by state ID.
    response = tator_api.get_media_list_by_id(project, {"state_ids": state_ids})
    assert len(response) == 1
    assert response[0].id == video

    # Test state retrival by media ID.
    response = tator_api.get_state_list_by_id(project, {"media_ids": [video]})
    assert len(response) == len(state_ids)

    # Test single create.
    state = random_state(project, state_type, video_obj, post=True)
    response = tator_api.create_state_list(project, body=state)
    assert isinstance(response, tator.models.CreateListResponse)
    print(response.message)
    state_id = response.id[0]
    response = tator_api.get_state(state_id)
    elemental_id = response.elemental_id
    version = response.version

    # Patch single state (in-place)
    patch = random_state(project, state_type, video_obj)
    patch_in_place = {**patch, "in_place": 1}
    response = tator_api.update_state(state_id, state_update=patch_in_place)
    assert "message" in response.to_dict()
    print(response.message)

    # Get single state.
    updated_state = tator_api.get_state_by_elemental_id(version, elemental_id)
    assert isinstance(updated_state, tator.models.State)
    assert_close_enough(patch, updated_state, exclude)

    # Get state by ID.
    state_by_id = tator_api.get_state_list_by_id(project, {"ids": [state_id]})
    assert len(state_by_id) == 1
    state_by_id = state_by_id[0]
    assert_close_enough(updated_state, state_by_id, exclude)

    # Patch single state.
    patch = random_state(project, state_type, video_obj)
    response = tator_api.update_state(state_id, state_update=patch)
    assert "message" in response.to_dict()
    print(response.message)

    # Get single state.
    updated_state = tator_api.get_state_by_elemental_id(version, elemental_id)
    assert isinstance(updated_state, tator.models.State)
    assert_close_enough(patch, updated_state, exclude)

    # Get state by ID.
    state_by_id = tator_api.get_state_list_by_id(project, {"ids": [updated_state.id]})
    assert len(state_by_id) == 1
    state_by_id = state_by_id[0]
    assert_close_enough(updated_state, state_by_id, exclude)

    # Delete single state.
    response = tator_api.delete_state_by_elemental_id(version, elemental_id)
    assert "message" in response.to_dict()
    print(response.message)

    params = {"media_id": [video], "type": state_type}
    assert tator_api.get_state_count(project, **params) == len(states)

    # Bulk update state attributes.
    bulk_patch = random_state(project, state_type, video_obj)
    bulk_patch = {"attributes": bulk_patch["attributes"], "in_place": 1}
    count = tator_api.get_state_count(project, **params)
    with pytest.raises(tator.openapi.tator_openapi.exceptions.ApiException):
        response = tator_api.update_state_list(project, **params, state_bulk_update=bulk_patch, count=count+1)
    response = tator_api.update_state_list(project, **params, state_bulk_update=bulk_patch, count=count)
    assert isinstance(response, tator.models.MessageResponse)
    print(response.message)

    # Bulk update specified states by ID.
    id_bulk_patch = random_state(project, state_type, video_obj)
    update_ids = random.choices(state_ids, k=100)
    id_bulk_patch = {"attributes": id_bulk_patch["attributes"], "ids": update_ids, "in_place": 1}
    response = tator_api.update_state_list(project, **params, state_bulk_update=id_bulk_patch)
    assert isinstance(response, tator.models.MessageResponse)
    print(response.message)

    # Verify all states have been updated.
    states = tator_api.get_state_list(project, **params)
    dataframe = tator.util.to_dataframe(states)
    assert len(states) == len(dataframe)
    for state in states:
        if state.id in update_ids:
            assert_close_enough(id_bulk_patch, state, exclude)
        else:
            assert_close_enough(bulk_patch, state, exclude)

    # Clone states to same media.
    version_mapping = {version.id: version.id for version in tator_api.get_version_list(project)}
    generator = tator.util.clone_state_list(
        tator_api,
        {**params, "project": project},
        project,
        version_mapping,
        {video: video},
        {},
        {state_type: state_type},
        tator_api,
    )
    for num_created, num_total, response, id_map in generator:
        print(f"Created {num_created} of {num_total} states...")
    print(f"Finished creating {num_created} states!")
    assert tator_api.get_state_count(project, **params) == 2 * len(states)

    # Delete all states.
    count = tator_api.get_state_count(project, **params)
    with pytest.raises(tator.openapi.tator_openapi.exceptions.ApiException):
        response = tator_api.delete_state_list(project, **params, count=count+1)
    response = tator_api.delete_state_list(project, **params, count=count)
    assert isinstance(response, tator.models.MessageResponse)

    # Verify all states are gone.
    states = tator_api.get_state_list(project, **params)
    assert states == []



import logging
import os
import sys

import tator
from ._upload_file import _upload_file

logger = logging.getLogger(__name__)

def update_applet(
        host: str,
        token: str,
        applet_id:int,
        html_file: str=None,
        applet_name: str=None,
        categories: list=None,
        description: str=None) -> None:
    """ Updates an existing applet using the provided parameters

    :param host: Host URL
    :param token: User token used for connecting to the host
    :param project: Unique identifier of project associated with algorithm
    :param html_file: Local path to applet html file
    :param applet_name: Name of the applet
    :param categories: Optional categories the applet belongs to
    :param description: Optional applet description
    """

    # Create the interface
    tator_api = tator.get_api(host=host, token=token)
    applet = tator_api.get_applet(id=applet_id)
    project = applet.project
    update = {}

    if html_file is not None:
        # Upload the file
        for progress, upload_info in _upload_file(tator_api, project, path=html_file):
            pass
        upload_url = tator_api.get_download_info(project, {'keys': [upload_info.key]})[0].url
        logger.info("Applet HTML file uploaded")

        # Save the uploaded file using the save report endpoint
        spec = tator.models.GenericFileSpec(
            name=os.path.basename(html_file),
            upload_url=upload_url)
        response = tator_api.save_generic_file(project=project, generic_file_spec=spec)
        logger.info(response)
        logger.info("Applet HTML file saved to project")

        update["html_file"] = response.url

    if applet_name is not None:
        update["name"] = applet_name

    if categories is not None:
        update["categories"] = categories

    if description is not None:
        update["description"] = description

    response = tator_api.update_applet(id=applet_id, applet_spec=update)

    log_msg = f"Applet updated (response = {response})"
    logger.info(log_msg)




import logging
import os
import sys

import tator
from ._upload_file import _upload_file

logger = logging.getLogger(__name__)

def register_applet(
        host: str,
        token: str,
        project: int,
        html_file: str,
        applet_name: str,
        categories: list=[],
        description: str='') -> None:
    """ Registers a applet using the provided parameters

    This will upload the given html file, save it to tator, and the new
    server side URL will be used when registering the applet. This may change
    the filename of the uploaded file

    :param host: Host URL
    :param token: User token used for connecting to the host
    :param project: Unique identifier of project associated with algorithm
    :param html_file: Local path to applet html file
    :param applet_name: Name of the applet
    :param categories: Optional categories the applet belongs to
    :param description: Optional applet description
    """

    # Create the interface
    tator_api = tator.get_api(host=host, token=token)

    # Upload the file
    for progress, upload_info in _upload_file(tator_api, project, path=html_file):
        pass
    upload_url = tator_api.get_download_info(project, {'keys': [upload_info.key]})[0].url
    logger.info("Applet HTML file uploaded")

    # Save the uploaded file using the save report endpoint
    spec = tator.models.GenericFileSpec(
        name=os.path.basename(html_file),
        upload_url=upload_url)
    response = tator_api.save_generic_file(project=project, generic_file_spec=spec)
    logger.info(response)
    logger.info("Applet HTML file saved to project")

    # With the applet file saved, register the applet
    spec = tator.models.Applet(
        name=applet_name,
        project=project,
        description=description,
        html_file=response.url,
        categories=categories)

    response = tator_api.register_applet(project=project, applet_spec=spec)

    log_msg = f"Applet registered with ID: {response.id}"
    logger.info(log_msg)



# coding: utf-8

"""
    Tator REST API

    Interface to the Tator backend.  # noqa: E501

    The version of the OpenAPI document: v1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from ..configuration import Configuration


class BucketOCINativeConfig(object):
    """
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'fingerprint': 'str',
        'key_content': 'str',
        'region': 'str',
        'tenancy': 'str',
        'user': 'str'
    }

    attribute_map = {
        'fingerprint': 'fingerprint',
        'key_content': 'key_content',
        'region': 'region',
        'tenancy': 'tenancy',
        'user': 'user'
    }

    def __init__(self, fingerprint=None, key_content=None, region=None, tenancy=None, user=None, local_vars_configuration=None):  # noqa: E501
        """BucketOCINativeConfig - a model defined in OpenAPI"""
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._fingerprint = None
        self._key_content = None
        self._region = None
        self._tenancy = None
        self._user = None
        self.discriminator = None

        self.fingerprint = fingerprint
        self.key_content = key_content
        self.region = region
        self.tenancy = tenancy
        self.user = user

    @property
    def fingerprint(self):
        """
        Public key fingerprint.

        :return: The fingerprint of this BucketOCINativeConfig. 
        :rtype: str
        """
        return self._fingerprint

    @fingerprint.setter
    def fingerprint(self, fingerprint):
        """
        Public key fingerprint.

        :param fingerprint: The fingerprint of this BucketOCINativeConfig.
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and fingerprint is None:  # noqa: E501
            raise ValueError("Invalid value for `fingerprint`, must not be `None`")  # noqa: E501

        self._fingerprint = fingerprint

    @property
    def key_content(self):
        """
        Private key content.

        :return: The key_content of this BucketOCINativeConfig. 
        :rtype: str
        """
        return self._key_content

    @key_content.setter
    def key_content(self, key_content):
        """
        Private key content.

        :param key_content: The key_content of this BucketOCINativeConfig.
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and key_content is None:  # noqa: E501
            raise ValueError("Invalid value for `key_content`, must not be `None`")  # noqa: E501

        self._key_content = key_content

    @property
    def region(self):
        """
        OCI region.

        :return: The region of this BucketOCINativeConfig. 
        :rtype: str
        """
        return self._region

    @region.setter
    def region(self, region):
        """
        OCI region.

        :param region: The region of this BucketOCINativeConfig.
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and region is None:  # noqa: E501
            raise ValueError("Invalid value for `region`, must not be `None`")  # noqa: E501

        self._region = region

    @property
    def tenancy(self):
        """
        Tenancy OCID.

        :return: The tenancy of this BucketOCINativeConfig. 
        :rtype: str
        """
        return self._tenancy

    @tenancy.setter
    def tenancy(self, tenancy):
        """
        Tenancy OCID.

        :param tenancy: The tenancy of this BucketOCINativeConfig.
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and tenancy is None:  # noqa: E501
            raise ValueError("Invalid value for `tenancy`, must not be `None`")  # noqa: E501

        self._tenancy = tenancy

    @property
    def user(self):
        """
        User OCID.

        :return: The user of this BucketOCINativeConfig. 
        :rtype: str
        """
        return self._user

    @user.setter
    def user(self, user):
        """
        User OCID.

        :param user: The user of this BucketOCINativeConfig.
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and user is None:  # noqa: E501
            raise ValueError("Invalid value for `user`, must not be `None`")  # noqa: E501

        self._user = user

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, BucketOCINativeConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, BucketOCINativeConfig):
            return True

        return self.to_dict() != other.to_dict()



#!/usr/bin/env python

import subprocess
import json
import logging
import os

from ..openapi.tator_openapi.models import MediaType
from ..openapi.tator_openapi.models import MessageResponse
from ..util.get_api import get_api
from ..util.get_parser import get_parser
from .transcode import find_best_encoder, get_length_info
from collections import defaultdict

STREAMING_RESOLUTIONS=[144, 360, 480, 720, 1080]

logger = logging.getLogger(__name__)

def parse_args():
    parser = get_parser()
    parser.add_argument('path', help='Path to original file.')
    parser.add_argument('--project', type=int, help='Unique integer identifying a project. This is '
                                                    'only needed if media_type is -1.')
    parser.add_argument('--media_type', type=int, help='Unique integer identifying a media type.')
    parser.add_argument('--group_to', type=int, default=480,
                         help='Vertical resolutions below this will be transcoded with '
                              'multi-headed ffmpeg.')
    parser.add_argument('--output', help='Path to output json file.')
    parser.add_argument('--media-id', type=int, help="Existing media ID, if applicable", default=-1)
    return parser.parse_args()

def _get_video_info(path):
    slow_cmd = [
        "ffprobe",
        "-v","error",
        "-show_entries", "stream:format=duration",
        "-print_format", "json",
        "-count_frames",
        "-skip_frame", "nokey",
        path,
    ]
    # The magic to reliably getting duration appears to be
    # the 'format=duration'. Out of an abundance of caution
    # we will leave the slow command in there in the event duration
    # slips by this faster method.
    fast_cmd = [
        "ffprobe",
        "-v","error",
        "-show_entries", "stream:format=duration",
        "-print_format", "json",
        path,
    ]
    output = subprocess.run(fast_cmd, stdout=subprocess.PIPE, check=True).stdout
    video_info = json.loads(output)
    if video_info['streams'][0].get('duration', None) is None:
        logger.info("NOTICE: Duration was null have to resort to slow method.")
        output = subprocess.run(slow_cmd, stdout=subprocess.PIPE, check=True).stdout
        video_info = json.loads(output)
    stream_idx=0
    audio=False
    for idx, stream in enumerate(video_info["streams"]):
        if stream["codec_type"] == "video":
            stream_idx=idx
        if stream["codec_type"] == "audio":
            logger.info("Found Audio Track")
            audio=True
    stream = video_info["streams"][stream_idx]
    stream = {**video_info.get("format", {}), **stream}
    fps, num_frames = get_length_info(stream)

    # Handle up to but not exceeding FHD
    height = int(stream["height"])
    width = int(stream["width"])

    # Handle rotated videos
    if "tags" in stream:
        if "rotate" in stream["tags"]:
            if stream["tags"]["rotate"] == "90":
                height = int(stream["width"])
                width = int(stream["height"])
    logger.info(f"Height of video is : {height}")
    return fps, num_frames, width, height, audio

def update_media(host, token, media_type, media_id, path):
    fps, num_frames, width, height, _ = _get_video_info(path)
    # Update the media object.
    api = get_api(host, token)
    response = api.update_media(media_id, media_update={
        'num_frames': num_frames,
        'fps': fps,
        'width': width,
        'height': height,
    })
    assert isinstance(response, MessageResponse)

def determine_transcode(host, token, media_type, media_id, path, group_to):
    """ Determine transcode workloads to be performed.

    :param host: URL of host.
    :param token: API token.
    :param media_type: Unique integer identifying a media type.
    :param media_id: Unique integer identifying the media object for which the transcode workloads
                     need to be determined.
    :param path: Path to original file.
    :param group_to: Resolutions less or equal to this will be grouped into one workload.
    """
    fps, num_frames, width, height, audio = _get_video_info(path)
    api = get_api(host, token)

    # Generate output path
    base, ext = os.path.splitext(path)

    # Get media type object.
    if media_type == -1:
        media_types = api.get_media_type_list(args.project)
        for media_type_obj in media_types:
            if media_type_obj.dtype == 'video':
                break
    else:
        media_type_obj = api.get_media_type(media_type)

    assert isinstance(media_type_obj, MediaType)

    # Get media object
    try:
        media_obj = api.get_media(media_id)
    except:
        media_obj = None

    if media_obj:
        # Get existing streaming/archival/audio file info
        if media_obj.media_files and media_obj.media_files.streaming:
            existing_streaming_resolutions = set(
                [
                    (stream_info.resolution[0], find_best_encoder(stream_info.codec))
                    for stream_info in media_obj.media_files.streaming
                ]
            )
        else:
            existing_streaming_resolutions = []

        if media_obj.media_files and media_obj.media_files.archival:
            existing_archival_resolutions = set(
                [
                    (archival_info.resolution[0], find_best_encoder(archival_info.codec))
                    for archival_info in media_obj.media_files.archival
                ]
            )
        else:
            existing_archival_resolutions = []

        # If at least one audio track exists, don't retranscode it
        if audio and media_obj.media_files and media_obj.media_files.audio:
            audio = False
    else:
        existing_streaming_resolutions = set()
        existing_archival_resolutions = set()

    available_resolutions = STREAMING_RESOLUTIONS
    crf_map = defaultdict(lambda: 23)
    codec_map = defaultdict(lambda: 'libx264')
    preset_map = defaultdict(lambda: '')
    pixel_format_map = defaultdict(lambda: 'yuv420p')
    try:
        if media_type_obj.streaming_config:
            available_resolutions = []
            for config in media_type_obj.streaming_config:
                available_resolutions.append(config.resolution)
                crf_map[config.resolution] = config.crf
                codec_map[config.resolution] = config.vcodec
                preset_map[config.resolution] = config.preset
                pixel_format_map[config.resolution] = config.pixel_format
    except Exception as e:
        # Likely an old version of the server
        # TODO: Remove me someday
        logger.info(e)
        logger.info("Defaulting to STREAMING_RESOLUTIONS")
    logger.info(f"Selected Resolutions {available_resolutions}")

    # Returns `True` if the resolution is smaller than the source video and it does not exist
    # already in the media object
    def _is_resolution_needed(resolution):
        if resolution > height:
            return False
        if (resolution, find_best_encoder(codec_map[resolution])) in existing_streaming_resolutions:
            return False
        return True

    # Make a list of resolutions needed
    resolutions = [res for res in available_resolutions if _is_resolution_needed(res)]
    if height < max(available_resolutions) and not height in resolutions:
        smallest_higher_res = min(r for r in available_resolutions if r > height)

        if not (height, codec_map[smallest_higher_res]) in existing_streaming_resolutions:
            # copy personality form the nearest higher resolution
            crf_map[height] = crf_map[smallest_higher_res]
            codec_map[height] = codec_map[smallest_higher_res]
            preset_map[height] = preset_map[smallest_higher_res]
            pixel_format_map[height] = pixel_format_map[smallest_higher_res]
            resolutions.append(height)

    # Streaming workloads (lower res)
    if any(res <= group_to for res in resolutions):
        workloads = [
            {
                "category": "streaming",
                "path": path,
                "raw_height": height,
                "raw_width": width,
                "configs": [
                    f"{res}:{crf_map[res]}:{codec_map[res]}:{preset_map[res]}:{pixel_format_map[res]}"
                    for res in resolutions
                    if res <= group_to
                ],
            }
        ]
    else:
        workloads = []

    # Streaming workloads (higher res)
    workloads.extend(
        {
            "category": "streaming",
            "path": path,
            "raw_height": height,
            "raw_width": width,
            "configs": [
                f"{res}:{crf_map[res]}:{codec_map[res]}:{preset_map[res]}"
            ],
        }
        for res in resolutions
        if res > group_to
    )

    # Archival workloads
    # This will force all transcodes to run `convert_archival` on the raw video,
    # which will fetch the media type and take action based on the returned 
    # archive config.
    if not existing_archival_resolutions:
        workloads.append(
            {
                'category': 'archival',
                'path': path,
                'raw_height': height,
                'raw_width': width,
                'configs': [],
            }
        )

    # Audio workloads
    if audio:
        workloads += [{
            'category': 'audio',
            'path': path,
            'raw_height': height,
            'raw_width': width,
            'configs': [],
        }]

    # Assign a sequential numerical ID to each workload.
    for idx in range(len(workloads)):
        workloads[idx]['id'] = idx

    return workloads

if __name__ == '__main__':
    args = parse_args()
    workloads = determine_transcode(
        args.host, args.token, args.media_type, args.media_id, args.path, args.group_to
    )
    for workload in workloads:
        workload['configs'] = ','.join(workload['configs'])
    with open(args.output, 'w') as f:
        json.dump(workloads, f)
